Module KappaExtImpl.

(* ==================================================================
     Kappa-extension implementation
   ================================================================== *)

Load "LTS.Transformation".
Import Transformation.

(* ----------------------------
     Build the kappa function
   ---------------------------- *)

Definition kappa_fun (s : State 1) : Act := (hd dummy s) * 10.

Lemma KappaFun_injective : forall s s', kappa_fun s = kappa_fun s' -> s = s'.
Proof.
unfold kappa_fun.
intros.
destruct s, s'.
simpl in H.
pose proof (vec1_eq dummy lst0 lst1).
assert (hd dummy lst0 = hd dummy lst1).
intuition.
apply listn_eq.
intuition.
Qed.

Lemma KappaFun_not_tau : forall s, kappa_fun s <> tau.
Proof.
unfold tau.
unfold kappa_fun.
intros.
intuition.
Qed.

Definition KappaFun := 
  Build_KFun kappa_fun KappaFun_injective KappaFun_not_tau.

(* --------------------------------
     Build the kappa extended LTS
   -------------------------------- *)

Definition KExtActs (G : LTS)(a : Act) := 
     Acts G a 
  \/ exists s, Init G s /\ kappa_fun s = a.

Definition KExtTrans (G : LTS)(a : Act)(s s' : State 1) :=
     Trans G a s s'
  \/ s = s' /\ Init G s /\ kappa_fun s = a.

Lemma KExt_trans_cond : 
  forall G a s s', KExtTrans G a s s' -> States G s /\ States G s' /\ KExtActs G a.
Proof.
unfold KExtTrans, KExtActs.
intros.
pose proof (trans_cond G a s s').
pose proof (init_cond G s).
intuition.
rewrite <- H ; assumption.
right ; exists s ; auto.
Qed.

Lemma KExt_acts_cond : 
  forall G a, KExtActs G a <-> exists s s', KExtTrans G a s s'.
Proof.
unfold KExtActs, KExtTrans.
intuition.
apply (acts_cond G a) in H0.
destruct H0 as [s [s' H0]].
exists s, s' ; left ; exact H0.
destruct H0 as [s H0].
exists s, s ; right ; intuition.
destruct H as [s [s' [H | H]]].
left ; pose proof (trans_cond G _ _ _ H) ; intuition.
right ; exists s ; intuition.
Qed.

Definition KExtLTS (G : LTS) := 
  Build_LTS 1 (States G) (KExtActs G) (KExtTrans G) (Init G)
              (KExt_trans_cond G) (init_cond G) 
              (init_not_empty_cond G) (KExt_acts_cond G).

(* ------------------------------------------------
     Build the kappa extended transformation rule
   ------------------------------------------------ *)

Lemma KExt_glue_cond : 
  forall r : TransRule,
    Init (KExtLTS (L r)) = Init (KExtLTS (R r)) /\
    Init (KExtLTS (L r)) = IntersectS (States (KExtLTS (L r))) (States (KExtLTS (R r))).
Proof.
simpl.
apply glue_cond.
Qed.

Definition KExtTransRule (r : TransRule) :=
  Build_TransRule (KExtLTS (L r)) (KExtLTS (R r)) (KExt_glue_cond r).

(* -------------------------------------------------------------------------
     The extended transformation rule adheres the TransRuleK specification
   ------------------------------------------------------------------------- *)

Axiom KappaIsReserved : 
  forall G H, H = KExtLTS G -> forall a, Acts G a -> ~(exists k, k * 10 = a).

Lemma KExt_kact_cond :
  forall r s, ~Acts (L r) (KappaFun s) /\ ~Acts (R r) (KappaFun s).
Proof.
simpl.
intros.
split ; 
  [ pose proof (KappaIsReserved (L r) (KExtLTS (L r))) 
  | pose proof (KappaIsReserved (R r) (KExtLTS (R r))) ] ;
intuition ;
specialize H1 with (KappaFun s) ;
intuition ; unfold kappa_fun in H ;
apply H ; exists (hd dummy s) ; intuition.
Qed.

Lemma KExt_s_cond : 
  forall r, let ker := KExtTransRule r in
    States (L ker) = States (L r) /\ States (R ker) = States (R r).
Proof.
simpl.
intuition.
Qed.

Lemma KExt_i_cond : 
  forall r, let ker := KExtTransRule r in
    Init (L ker) = Init (L r) /\ Init (R ker) = Init (R r).
simpl.
intuition.
Qed.

Lemma KExt_a_cond : 
  forall r a, let ker := KExtTransRule r in
        (Acts (L ker) a -> Acts (L r) a \/ exists s, Init (L r) s /\ KappaFun s = a)
     /\ (Acts (L r) a -> Acts (L ker) a)
     /\ (Acts (R ker) a -> Acts (R r) a \/ exists s, Init (R r) s /\ KappaFun s = a)
     /\ (Acts (R r) a -> Acts (R ker) a).
Proof.
simpl.
unfold KExtActs.
intuition.
Qed.

Lemma KExt_t_cond : 
  forall r a s s', let ker := KExtTransRule r in
       (Trans (L ker) a s s' <-> Trans (L r) a s s'
                             \/ (KappaFun s = a /\ s = s' /\ Init (L r) s))
    /\ (Trans (R ker) a s s' <-> Trans (R r) a s s' 
                             \/ (KappaFun s = a /\ s = s' /\ Init (R r) s)).
Proof.
simpl.
unfold KExtTrans.
intros.
intuition.
Qed.

Definition TransRuleKImpl (r : TransRule) := 
  Build_TransRuleK r KappaFun (KExtTransRule r) (KExt_kact_cond r) 
    (KExt_s_cond r) (KExt_i_cond r) (KExt_a_cond r) (KExt_t_cond r).

End KappaExtImpl.