(* Basic notions for LTS (LTS, BB) *)
Module LTSN.

Require Export Relations (* relative-transitive-closure *).

Load "Utils.ListN".
Export ListN.
Load "Utils.SetTheoryProps".
Export SetTheoryProps.

Definition Act := nat (* An action is identified with a string. *).
Definition tau := 1  (* The silent action *).

Notation state_type := nat.
Definition State (n : nat) := listn state_type n (* A state is identified with 'n' naturals. *).

Definition IntersectS {n} := Intersection (State n).
Definition SetminusS {n}  := Setminus (State n).
Definition UnionS {n}     := Union (State n).
Definition SingletonS {n} := Singleton (State n).

(* ==================================================================
     LTS
   ================================================================== *)

(* LTS *)
Record LTS {n : nat} := {
  States : Ensemble (State n); (* States *)
  Acts   : Ensemble Act;   (* Actions *)
  Trans  : Act -> (State n) -> (State n) -> Prop;
  Init   : Ensemble (State n);

  (* Transitions involve members of States and labels in Acts. *)
  trans_cond : forall a s s', Trans a s s' -> States s /\ States s' /\ Acts a;

  (* The initial states are member of the States of the LTS *)
  init_cond : forall s, Init s -> States s;

  (* The set of initial state may not be empty *)
  init_not_empty_cond : exists s, Init s;

  (* For every action in Acts, there is a transition *)
  acts_cond : forall a, Acts a -> exists s s', Trans a s s'
}.

Lemma states_not_empty : forall {n} G, exists s : State n, States G s.
Proof.
intros.
destruct (init_not_empty_cond G).
exists x.
apply init_cond.
exact H.
Qed.

(* The reflexive transitive closure of the tau-transition: --tau--> 
   (A sequence of (connected) tau-transitions of length at least 0) *)
Definition TauPath {n}(G : LTS) := clos_refl_trans (State n) (Trans G tau).

(* Branching Bisimulation (part of Def. 3 in [1]) *)
Record BB {n}(G1 : LTS)(G2 : LTS) := {
  BR :> (State n) -> (State n) -> Prop;

  (* B relates states of G1 to states of G2 *)
  B_states : forall s t, BR s t -> States G1 s /\ States G2 t;

  (* The initial states of both LTSs must be related *)
  B_init : forall s, (Init G1 s -> exists t, Init G2 t /\ BR s t)
                  /\ (Init G2 s -> exists t, Init G1 t /\ BR t s);
  
  (* B is a branching bisimulation *)
  B_branching : 
    forall s t, BR s t ->
      forall a s', Trans G1 a s s' -> 
           (a = tau /\ BR s' t) 
        \/ (exists th t', (TauPath G2 t th /\ Trans G2 a th t' /\ BR s th /\ BR s' t'));

  B_branching_sym :
    forall s t, BR s t ->
      forall a t', Trans G2 a t t' -> 
           (a = tau /\ BR s t') 
        \/ (exists sh s', (TauPath G1 s sh /\ Trans G1 a sh s' /\ BR sh t /\ BR s' t'))
}.


(* ==================================================================
     Dummy LTS
   ================================================================== *)

(* The Singleton LTS *)
Parameter dummy : nat (* The one state in the LTS *).
Definition d_state := Bld_list1 dummy.
Definition d_staten (n : nat) := Bld_listn (alist n dummy) (alist_length n dummy).
Definition No_Trans (a : Act)(s1 s2 : State 1) : Prop := False.

Lemma d_staten_nth : forall n i, nth i (d_staten n) dummy = dummy.
Proof.
intros.
unfold d_staten.
simpl.
rewrite alist_nth.
reflexivity.
Qed.

Lemma Singleton_LTS_trans_cond :
  forall a s s', No_Trans a s s' -> SingletonS d_state s /\ SingletonS d_state s' /\ Empty_set Act a.
Proof.
intuition.
Qed.

Lemma Singleton_LTS_init_cond :
  forall s, SingletonS d_state s -> SingletonS d_state s.
Proof.
intros; exact H.
Qed.

Lemma Singleton_LTS_init_not_empty_cond :
  exists s, SingletonS d_state s.
Proof.
exists d_state.
unfold SingletonS.
apply In_singleton.
Qed.

Lemma Singleton_LTS_acts_cond :
  forall a, Empty_set Act a -> exists s s', No_Trans a s s'.
Proof.
intuition.
Qed.

(* We take the full set of actions to support arbitrary 
   synchronisation laws in transformation networks defined later *)
Definition Singleton_LTS := 
  Build_LTS 1 (SingletonS d_state) (Empty_set Act) (No_Trans) (SingletonS d_state)
              (Singleton_LTS_trans_cond) (Singleton_LTS_init_cond) 
              (Singleton_LTS_init_not_empty_cond) (Singleton_LTS_acts_cond).

End LTSN.