(* Network of LTSs and related notions *)
Module Network.

Load "Utils.ClassicalExt".
Import ClassicalExt.
Load "Utils.OptionProps".
Import OptionProps.
Load "LTS.Transformation".
Export Transformation.

(* ==================================================================
     LTS Network
   ================================================================== *)

(* Syncronisation law *)
Record SyncLaw {n} := {
  sync : listn (option Act) n; (* Synchronisation vector, * is represented by None *)
  act  : Act;                  (* Resulting action *)

  sync_cond : exists i, nth i sync None <> None
}.

(* Act membership for elements of a synchronisation vector *)
Definition OActs {n}(G : @LTS n)(a : option Act) :=
  match a with
    | None => False      (* a is dummy action * *)
    | Some x => Acts G x
  end.

(* Act U {*} membership for elements of a synchronisation vector *)
Definition DActs {n}(G : @LTS n)(a : option Act) :=
  match a with
    | None => True      (* a is dummy action * *)
    | Some x => Acts G x
  end.

Lemma OActs_In_to_Some : forall {n}(G : @LTS n) a, OActs G a -> exists b, a = Some b /\ Acts G b.
Proof.
unfold OActs.
intros.
destruct a.
exists a ; intuition.
contradiction.
Qed.

Definition NoSyncAndRename {n}(V : Ensemble (@SyncLaw n)) := 
  forall i v, V v /\ i < n /\ (nth i (sync v) None) = Some tau -> 
    (act v) = tau /\ forall j, j < n -> (nth j (sync v) None) <> None -> i = j.

Definition NoCut {n}(V : Ensemble (@SyncLaw n))(Glist : list LTS) :=
  forall i, (Acts (nth i Glist d_lts) tau) -> exists v, V v /\ (nth i (sync v) None) = Some tau.

Definition Admissible {n}(V : Ensemble (@SyncLaw n))(Glist : list LTS) :=
 NoSyncAndRename V /\ NoCut V Glist.

(* LTS network *)
Record LTSNetwork {n} := {
  Procs : listn (@LTS 1) n;      (* Vector of LTSs (with state size 1) *)
  SyncN : Ensemble (@SyncLaw n); (* Set of synchronisation rules *)

  Adm_cond      : Admissible SyncN Procs
}.

(* ==================================================================
     The system LTS of an LTS network
   ================================================================== *)

(* Convert the i-th element of a State of size n to a (State 1) *)
Definition nthS {n}(i : nat)(s : State n) : State 1 := 
  Bld_list1 (nth i s dummy).

(* A n-vector state is in the system LTS iff
   for each i-th element in the state vector 
   the element is in the i-th LTS *)
Definition NetStates {n}(M : @LTSNetwork n)(s : State n) :=
  forall i, i < n -> States (nth i (Procs M) d_lts) (nthS i s).

(* Transition with known sync law *)
Definition NetTransSL {n}(M : @LTSNetwork n)(v : SyncLaw)(s1 s2 : State n) := 
  SyncN M v /\
    forall (i : nat), i < n -> (* forall i in 1 .. n *)
      match nth i (sync v) None with
        | None =>    nth i s1 dummy = nth i s2 dummy 
                  /\ States (nth i (Procs M) d_lts) (nthS i s1)
                  /\ States (nth i (Procs M) d_lts) (nthS i s2)
        | Some x => Trans (nth i (Procs M) d_lts) x (nthS i s1) (nthS i s2)
      end.

(* Transition in a system LTS *)
Definition NetTrans {n}(M : LTSNetwork)(a : Act)(s1 s2 : State n) := 
  exists v, act v = a /\ NetTransSL M v s1 s2.

(* Actions of the system LTS *)
Definition NetActs {n}(M : @LTSNetwork n)(a : Act) :=
  exists s s', NetTrans M a s s'.

(* Initial states of the system LTS *)
Definition NetInit {n}(M : @LTSNetwork n)(s : State n) :=
  forall i, i < n -> Init (nth i (Procs M) d_lts) (nthS i s).

Lemma NetLTS_trans_cond :
  forall {n} M a (s s' : State n), NetTrans M a s s' -> NetStates M s /\ NetStates M s' /\ NetActs M a.
Proof.
intros.
intros.
destruct H as [l [H0 [H1 H2]]].
unfold NetStates, NetActs.
intuition ;
try (
  specialize H2 with i ;
  destruct (nth i (sync l) None) ;
  intuition ;
  pose proof (trans_cond (nth i (Procs M) d_lts) a0 (nthS i s) (nthS i s') H3) ;
  intuition ).
unfold NetActs.
exists s, s'.
exists l.
split ; try exact H0.
unfold NetTransSL.
intuition.
Qed.

Lemma NetLTS_init_cond :
  forall {n} M (s : State n), NetInit M s -> NetStates M s.
Proof.
unfold NetInit.
unfold NetStates.
intros.
specialize H with i.
pose proof (init_cond (nth i (Procs M) d_lts) (nthS i s)).
intuition.
Qed.
  
Lemma NetLTS_init_not_empty_cond :
  forall {n} M, exists (s : State n), NetInit M s.
Proof.
intros.
unfold NetInit.
unfold nthS.
set (InitProp := fun i s => Init (nth i (Procs M) d_lts) (Bld_list1 s)).
pose proof (listn_forall_exists _ n dummy InitProp).
unfold InitProp in H.
apply H ; clear H.
intros.
destruct (init_not_empty_cond (nth i (Procs M) d_lts)) as [s Hs].
exists (hd dummy s).
rewrite list1_prop.
exact Hs.
Qed.

Lemma NetLTS_acts_cond : forall {n} (M : @LTSNetwork n), 
  forall a, NetActs M a -> exists s s', NetTrans M a s s'.
Proof.
unfold NetActs.
intuition.
Qed.

(* The System LTS *)
Definition SystemLTS {n}(M : LTSNetwork) := 
  Build_LTS n (NetStates M) (NetActs M) (NetTrans M) (NetInit M)
              (NetLTS_trans_cond M) (NetLTS_init_cond M) 
              (NetLTS_init_not_empty_cond M) (NetLTS_acts_cond M).

(* The active set of a synchronisation law *)
Definition Ac {n}(v : @SyncLaw n)(i : nat) := exists a, nth i (sync v) None = Some a (* implicit i < n *).

(* The set of syncronising actions *)
Definition As {n}(v : @SyncLaw n) a := 
    (exists i, nth i (sync v) None = Some a /\ 
      (exists j, Ac v j /\ i <> j)) (* |Ac v| > 1 *).

Lemma AcSome_prop : 
  forall {n} (v : @SyncLaw n) a i, nth i (sync v) None = Some a -> (Ac v) i.
Proof.
unfold Ac.
intros.
exists a.
exact H.
Qed.

Lemma Acv_i_st_n : forall {n} (v : @SyncLaw n) i, Ac v i -> i < n.
Proof.
intros.
destruct H.
apply (listn_nth_bound None (sync v)).
congruence.
Qed.

Lemma SyncLawNone_to_NotAc : 
  forall {n} (v : @SyncLaw n) i, nth i (sync v) None = None -> ~Ac v i.
Proof.
unfold not.
intuition.
destruct H0.
congruence.
Qed.

Lemma NotAc_to_SyncLawNone :
  forall {n} (v : @SyncLaw n) i, ~Ac v i -> nth i (sync v) None = None.
Proof.
unfold not.
intuition.
unfold Ac in H.
apply NNPP.
unfold not ; intros.
apply H.
destruct (OptionSum (nth i (sync v) None)) as [HO | HO].
destruct HO.
exists x.
exact e.
contradiction.
Qed.

Lemma AcNone_prop : 
  forall {n} (v : @SyncLaw n) i, nth i (sync v) None = None <-> ~Ac v i.
Proof.
intros ; split.
apply SyncLawNone_to_NotAc.
apply NotAc_to_SyncLawNone.
Qed.

Lemma IndependentAction_prop : forall {n} (M : @LTSNetwork n) v a,
  SyncN M v -> ~As v a -> 
    forall i j, nth i (sync v) None = Some a -> Ac v j -> i = j.
Proof.
intros.
apply NNPP ; intuition.
apply H0.
unfold As.
exists i.
split ; try exact H1.
exists j.
auto.
Qed.

Lemma NetTransSL_i_not_in_Ac :
  forall {n} (M : @LTSNetwork n) v s s', NetTransSL M v s s' -> 
    forall i, i < n -> ~Ac v i -> nthS i s = nthS i s'.
Proof.
intros.
destruct H as [_ HT].
specialize HT with i.
apply AcNone_prop in H1.
rewrite H1 in HT.
unfold nthS.
apply list1_eq.
intuition.
Qed.

Lemma NetTranSL_or_introl : 
  forall {n} (M : @LTSNetwork n) v s s', NetTransSL M v s s' -> 
    NetTransSL M v s s' \/ NetTransSL M v s' s.
Proof.
intuition.
Qed.

Lemma NetTranSL_or_intror : 
  forall {n} (M : @LTSNetwork n) v s s', NetTransSL M v s' s -> 
    NetTransSL M v s s' \/ NetTransSL M v s' s.
Proof.
intuition.
Qed.

(* ==================================================================
     Conversion from (Ensemble nat) to (list bool)
   ================================================================== *)

Axiom EnsembleNat_dec : forall (I : Ensemble nat) i, {I i} + {~I i}.

Definition In2bool (I : Ensemble nat)(i : nat) :=
  if EnsembleNat_dec I i then true else false.

(* Build bool-list with index-range [0,m) (with "None" when out of range) *)
Fixpoint Ensemble2boollist' (I : Ensemble nat)(m : nat) : list (option bool) :=
  match m with
  | 0    => nil
  | S m' => cons (Some (In2bool I m')) (Ensemble2boollist' I m')
  end.

Definition Ensemble2boollist (n : nat)(I : Ensemble nat) := rev (Ensemble2boollist' I n).

Lemma Ensemble2boollist'_length : forall I m, length (Ensemble2boollist' I m) = m.
Proof.
intros.
induction m0 ; simpl.
reflexivity.
intuition.
Qed.

Lemma Ensemble2boollist_length : forall n I, length (Ensemble2boollist n I) = n.
Proof.
unfold Ensemble2boollist.
intros.
rewrite rev_length.
apply Ensemble2boollist'_length.
Qed.

Lemma Ensemble2boollist'_to_In :
  forall (I : Ensemble nat) i m, nth i (Ensemble2boollist' I m) None = Some true -> I (m - S i).
Proof.
intros.
unfold Ensemble2boollist', In2bool in H.
dependent induction i.
(* Base case *)
destruct m0.
simpl in H.
congruence.
simpl in *.
inversion H.
assert (m0 - 0 = m0) by omega.
rewrite H0.
case_eq (EnsembleNat_dec I m0) ; intros.
exact i.
rewrite H2 in H1 ; intuition.
(* Step case *)
dependent destruction m0.
simpl in *.
congruence.
simpl.
apply IHi.
intuition.
Qed.

Lemma Ensemble2boollist'_from_In :
  forall (I : Ensemble nat) i m, i < m -> I (m - S i) -> nth i (Ensemble2boollist' I m) None = Some true.
Proof.
intros.
unfold Ensemble2boollist', In2bool.
dependent induction i.
(* Base case *)
destruct m0.
intuition.
simpl in *.
assert (m0 - 0 = m0) by omega.
rewrite H1 in H0.
case_eq (EnsembleNat_dec I m0) ; intros.
reflexivity.
contradiction.
(* Step case *)
dependent destruction m0.
simpl in *.
intuition.
simpl in *.
apply IHi ;
intuition.
Qed.

Lemma Ensemble2boollist'_to_not_In :
  forall (I : Ensemble nat) i m, nth i (Ensemble2boollist' I m) None = Some false -> ~I (m - S i).
Proof.
intros.
unfold not; intro.
unfold Ensemble2boollist', In2bool in H.
dependent induction i.
(* Base case *)
destruct m0 ; simpl in *.
congruence.
inversion H.
assert (m0 - 0 = m0) by omega.
rewrite H1 in H0.
case_eq (EnsembleNat_dec I m0) ; intros ; try rewrite H3 in H2 ; congruence.
(* Step case *)
dependent destruction m0.
simpl in H ; congruence.
simpl in *.
apply (IHi m0) ; intuition.
Qed.

Lemma Ensemble2boollist'_from_not_In :
  forall (I : Ensemble nat) i m, i < m -> ~I (m - S i) -> nth i (Ensemble2boollist' I m) None = Some false.
Proof.
intros.
unfold Ensemble2boollist', In2bool, not in *.
dependent induction i.
(* Base case *)
destruct m0 ; simpl in *.
intuition.
assert (m0 - 0 = m0) by omega.
rewrite H1 in H0.
case_eq (EnsembleNat_dec I m0) ; intros.
contradiction.
reflexivity. 
(* Step case *)
dependent destruction m0.
intuition.
simpl in *.
apply IHi.
intuition.
exact H0.
Qed.

Lemma Ensemble2boollist_to_not_In :
  forall n (I : Ensemble nat) i, nth i (Ensemble2boollist n I) None = Some false -> ~I i.
Proof.
intros.
pose proof (list_nth_bound _ _ _ (Option_Some_NNone _ _ H)).
rewrite (Ensemble2boollist_length n I) in H0.
unfold Ensemble2boollist in H.
pose proof (Ensemble2boollist'_length I n).
rewrite rev_nth, H1 in H.
remember (n - S i) as i'.
assert (i = n - S i').
intuition.
rewrite H2.
apply Ensemble2boollist'_to_not_In.
exact H.
intuition.
Qed.

Lemma Ensemble2boollist_to_In :
  forall n (I : Ensemble nat) i, nth i (Ensemble2boollist n I) None = Some true -> I i.
Proof.
intros.
pose proof (list_nth_bound _ _ _ (Option_Some_NNone _ _ H)).
rewrite (Ensemble2boollist_length n I) in H0.
unfold Ensemble2boollist in H.
pose proof (Ensemble2boollist'_length I n).
rewrite rev_nth, H1 in H.
remember (n - S i) as i'.
assert (i = n - S i').
intuition.
rewrite H2.
apply Ensemble2boollist'_to_In.
exact H.
intuition.
Qed.

Lemma Ensemble2boollist_from_In :
  forall n (I : Ensemble nat) i, i< n -> I i -> nth i (Ensemble2boollist n I) None = Some true.
Proof.
intros.
unfold Ensemble2boollist.
pose proof (Ensemble2boollist'_length I n).
rewrite rev_nth ; rewrite H1.
remember (n - S i) as i'.
apply Ensemble2boollist'_from_In.
intuition.
assert (i = n - S i') as Hrw; intuition ; rewrite <- Hrw.
exact H0.
exact H.
Qed.

(* Fore some cases the "i < n" limit may be removed.
For now we keep this for backwards compatibility for many proofs *)
Lemma Ensemble2boollist_In :
  forall n (I : Ensemble nat) i, i < n ->
    (    (nth i (Ensemble2boollist n I) None = Some true <-> I i)
      /\ (nth i (Ensemble2boollist n I) None = Some false <-> ~I i) ).
Proof.
unfold Ensemble2boollist.
intros.
pose proof (Ensemble2boollist'_length I n).
rewrite rev_nth ; rewrite H0.
remember (n - S i) as i'.
assert (i = n - S i').
intuition.
rewrite H1.
split ; split ; intros ;
  [ apply Ensemble2boollist'_to_In     | apply Ensemble2boollist'_from_In 
  | apply Ensemble2boollist'_to_not_In | apply Ensemble2boollist'_from_not_In ] ;
intuition.
exact H.
Qed.

(* ==================================================================
     Filtering an LTS Network
   ================================================================== *)

(* Filter for applicable SyncLaws *)
Fixpoint SyncLaw_applicable' (Iv : list (option bool * option Act)) : Prop :=
  match Iv with
  | nil                       => True
  | (Some true , _)      :: l => SyncLaw_applicable' l
  | (Some false, None)   :: l => SyncLaw_applicable' l
  | (Some false, Some a) :: l => False
  | (None      , _)      :: l => SyncLaw_applicable' l
  end.

Lemma SyncLaw_applicable'_i_not_in_I_None :
  forall (Iv : list (option bool * option Act)), 
    SyncLaw_applicable' Iv -> forall i, fst (nth i Iv (None, None)) = Some false -> 
                                          snd (nth i Iv (None, None)) = None.
Proof.
intros.
unfold SyncLaw_applicable' in H.
dependent induction Iv.
  (* Base *)
  destruct i ; simpl ; reflexivity.
  (* Step *)
  destruct i.
    (* i = 0 *)
    destruct a.
    simpl in H0 |- *.
    rewrite H0 in H.
    destruct o0.
    contradiction.
    reflexivity.
    (* i = S i *)
    simpl in *.
    apply IHIv ; try exact H0.
    destruct a.
    simpl in *.
    destruct o.
    destruct b ; try exact H.
    destruct o0.
    contradiction.
    exact H.
    exact H.
Qed.

Lemma SyncLaw_applicable'_all_i_not_in_I_None :
  forall (Iv : list (option bool * option Act)), 
    (forall i, i < length Iv -> fst (nth i Iv (None, None)) = Some false -> 
       snd (nth i Iv (None, None)) = None) -> SyncLaw_applicable' Iv.
Proof.
intros.
unfold SyncLaw_applicable'.
dependent induction Iv.
  (* Base *)
  trivial.
  (* Step *)
  destruct a.
  destruct o.
  destruct b.
    (* b = true *)
    apply IHIv.
    intros.
    apply (H (S i)).
    simpl.
    omega.
    simpl.
    exact H1.
    (* b = false *)
    pose proof (H 0).
    simpl in H0.
    rewrite H0 ; try reflexivity.
    apply IHIv.
    intros.
    apply (H (S i)).
    simpl.
    omega.
    simpl.
    exact H2.
  omega.
  apply IHIv.
  intros.
  apply (H (S i)) ;
  simpl.
  omega.
  exact H1.
Qed.

Definition SyncLaw_applicable {n}(I : Ensemble nat)(v : @SyncLaw n) : Prop := 
  SyncLaw_applicable' (combine (Ensemble2boollist n I) (sync v)).

Lemma SyncLaw_applicable_i_not_in_I_None :
  forall {n}(I : Ensemble nat)(v : @SyncLaw n), 
    SyncLaw_applicable I v -> forall i, i < n -> ~I i ->  nth i (sync v) None = None.
Proof.
intros.
unfold SyncLaw_applicable in H.
apply (SyncLaw_applicable'_i_not_in_I_None) with (i := i) in H.
rewrite combine_nth in H.
simpl in H.
exact H.
rewrite (Ensemble2boollist_length n I).
rewrite (lst_length _ _ (sync v)).
reflexivity.
rewrite combine_nth.
simpl.
pose proof (Ensemble2boollist_In n I i) ; intuition.
rewrite (Ensemble2boollist_length n I).
rewrite (lst_length _ _ (sync v)).
reflexivity.
Qed.

Lemma SyncLaw_applicable_all_i_not_in_I_None :
  forall {n}(I : Ensemble nat)(v : @SyncLaw n), 
    (forall i, i < n -> ~I i -> nth i (sync v) None = None) -> SyncLaw_applicable I v.
Proof.
intros.
unfold SyncLaw_applicable.
apply (SyncLaw_applicable'_all_i_not_in_I_None).
intros.
rewrite combine_length in H0.
rewrite combine_nth in * ;
rewrite (Ensemble2boollist_length n I) in * ; 
rewrite (lst_length _ _ (sync v)) in * ; try reflexivity.
simpl in *.
specialize H with i.
apply (Ensemble2boollist_to_not_In n I i) in H1.
assert (min n n = n) by intuition.
rewrite H2 in H0.
intuition.
Qed.

Definition SyncF {n}(M : @LTSNetwork n)(I : Ensemble nat)(v : SyncLaw) := SyncN M v /\ SyncLaw_applicable I v.

Lemma SyncF_to_SyncN : forall {n}(M : @LTSNetwork n)(I : Ensemble nat) v, SyncF M I v -> SyncN M v.
Proof.
unfold SyncF ; intuition.
Qed.

Lemma SyncF_i_not_in_I_None :
  forall {n}(M : @LTSNetwork n)(I : Ensemble nat)(v : SyncLaw), 
    SyncF M I v -> forall i, i < n -> ~I i -> nth i (sync v) None = None.
Proof.
intros.
unfold SyncF in H.
destruct H as [Hv HFv].
apply (SyncLaw_applicable_i_not_in_I_None I v) with (i := i) in HFv ; assumption.
Qed.

Lemma SyncF_all_i_not_in_I_None :
  forall {n}(M : @LTSNetwork n)(I : Ensemble nat)(v : SyncLaw), 
    (forall i, i < n -> ~I i -> nth i (sync v) None = None) /\ SyncN M v -> SyncF M I v.
Proof.
intros.
unfold SyncF.
destruct H.
split.
exact H0.
apply SyncLaw_applicable_all_i_not_in_I_None.
exact H.
Qed.

(* Filter LTSs *)
   
Definition Filter' (bG : option bool * LTS) :=
  match bG with
  | (None, _) => d_lts
  | (Some false, _) => d_lts
  | (Some true, G) => G 
  end.

Definition Filter {n}(l : listn LTS n)(I : Ensemble nat) := 
  map Filter' (combine (Ensemble2boollist n I) l).

Lemma Filter_length : forall {n} (l : listn LTS n) I, length (Filter l I) = n.
Proof.
intros.
unfold Filter.
rewrite map_length.
rewrite combine_length.
rewrite Ensemble2boollist_length.
rewrite lst_length.
intuition.
Qed.

Definition FilterList {n}(l : listn LTS n)(I : Ensemble nat) := 
  Bld_listn (Filter l I) (Filter_length l I).

Lemma Filter_nth :
  forall {n} (l : listn LTS n) I i,
    nth i (Filter l I) d_lts = match (nth i (Ensemble2boollist n I) None) with
                               | None => d_lts
                               | Some false => d_lts
                               | Some true => (nth i l d_lts) end.
Proof.
intros.
unfold Filter.
pose proof (map_nth (Filter') (combine (Ensemble2boollist n I) l) (None, d_lts) i).
unfold Filter' in *.
simpl in H.
rewrite H.
rewrite combine_nth.
reflexivity.
rewrite Ensemble2boollist_length.
rewrite lst_length.
reflexivity.
Qed.

Lemma Filter_In : 
  forall {n} (l : listn LTS n) (I : Ensemble nat) i, i < n ->
       (I i -> nth i (FilterList l I) d_lts = nth i l d_lts)
    /\ (~I i -> nth i (FilterList l I) d_lts = d_lts).
Proof.
unfold FilterList.
unfold Bld_listn.
simpl.
intros.
rewrite Filter_nth.
destruct (Ensemble2boollist_In n I i H) as [HIn HNIn].
split ; intro ;
 [ apply HIn in H0 | apply HNIn in H0 ] ;
rewrite H0 ;
reflexivity.
Qed.

Lemma FilterAdm_cond : 
  forall {n} (M : @LTSNetwork n) I, Admissible (SyncF M I) (FilterList (Procs M) I).
Proof.
simpl.
intros.
destruct (Adm_cond M).
unfold Admissible in *.
split.
(* NoSyncAndRename *)
unfold NoSyncAndRename, SyncF in H |- *.
intros.
specialize H with i v.
intuition.
(* NoCut *)
unfold NoCut in *.
intros.
rewrite Filter_nth in H1.
case_eq (nth i (Ensemble2boollist n I) None) ; [intros b HI | intro HI] ; try destruct b ; rewrite HI in H1 ; simpl in H1.
    (* I i *)
    destruct (H0 i H1) as [v [Hv Hvt]].
    exists v.
    split ; try exact Hvt.
    apply (SyncF_all_i_not_in_I_None M I v).
    split ; try exact Hv.
    pose proof (Acv_i_st_n _ _ (AcSome_prop _ _ _ Hvt)).  
    intros.
    apply NNPP.
    unfold not ; intro Hvi0.
    pose proof (listn_nth_bound _ _ _ Hvi0).
    apply H4.
    apply (Ensemble2boollist_In n I i H2) in HI.
    destruct (H i v) ; intuition.
    specialize H7 with i0 ; intuition.
    rewrite <- H7.
    exact HI.
    (* ~ I i *)
    contradiction H1.
  (* nth i (Ensemble2boollist n I) None = None *)
  contradiction.
Qed.

Definition FilterNetwork {n}(M : LTSNetwork)(I : Ensemble nat) := 
  Build_LTSNetwork n (FilterList (Procs M) I)
                     (SyncF M I)
                     (FilterAdm_cond M I).

Lemma Filter_States_eq :
  forall {n} M I (s1 s2 : State n), 
        States (SystemLTS (FilterNetwork M I)) s1
    /\  States (SystemLTS (FilterNetwork M I)) s2
    /\ (forall i, i < n /\ I i -> (nthS i s1) = (nthS i s2))
         -> s1 = s2.
Proof.
simpl.
unfold NetStates.
unfold nthS.
simpl.
intros.
destruct H as [H [H1 H2]].
destruct s1, s2.
simpl in *.
assert (lst0 = lst1).
apply (list_eq lst0 lst1 dummy).
intuition.
intros.
assert (i < n \/ n <= i) by omega.
destruct H0.
destruct (Filter_In (Procs M) I i H0) as [HIn HNIn].
specialize H2 with i.
destruct (EnsembleNat_dec I i).
intuition.
unfold Bld_list1 in *.
congruence.
apply HNIn in n0.
unfold FilterList in n0.
simpl in n0.
specialize H with i.
specialize H1 with i.
rewrite n0 in H, H1.
simpl in H, H1.
intuition.
apply Singleton_inv in H3.
apply Singleton_inv in H.
unfold d_state in H3, H.
unfold Bld_list1 in H3, H.
congruence.
rewrite nth_overflow ; intuition.
rewrite nth_overflow ; intuition.
assert (length lst0 = length lst1) by omega.
apply listn_eq.
exact H0.
Qed.

Lemma FilterNetwork_Procs_i_in_I : 
  forall {n} (M : @LTSNetwork n) (I : Ensemble nat) i, i < n -> I i ->
    (nth i (Procs (FilterNetwork M I)) d_lts) = (nth i (Procs M) d_lts).
Proof.
intros.
simpl.
rewrite Filter_nth.
assert (nth i (Ensemble2boollist n I) None = Some true) as HI by
  ( apply Ensemble2boollist_In ; assumption ).
rewrite HI.
reflexivity.
Qed.

Lemma FilterNetwork_Procs_i_not_in_I : 
  forall {n} (M : @LTSNetwork n) (I : Ensemble nat) i, i < n -> ~I i ->
    (nth i (Procs (FilterNetwork M I)) d_lts) = d_lts.
Proof.
intros.
simpl.
rewrite Filter_nth.
assert (nth i (Ensemble2boollist n I) None = Some false) as HI by
  ( apply Ensemble2boollist_In ; assumption ).
rewrite HI.
reflexivity.
Qed.

(* The active sets of transitions in a filtered network are included in I *)
Lemma FilterNetwork_Trans_Ac_in_I : 
  forall {n} (M : @LTSNetwork n) I v s s',
    NetTransSL (FilterNetwork M I) v s s' ->
      forall i, Ac v i -> I i.
Proof.
intros.
destruct H0 as [a Ha].
destruct H as [_ HT].
specialize HT with i.
rewrite Ha in HT.
assert (i < n) as Hin by ( apply (listn_nth_bound None (sync v)) ; congruence ).
simpl in HT.
rewrite Filter_nth in HT.
simpl in HT.
case_eq (nth i (Ensemble2boollist n I) None) ; [intros b HEI | intro HEI] ; rewrite HEI in HT ; 
  try (destruct b ; apply Ensemble2boollist_In in HEI ; try exact Hin) ; intuition.
Qed.

Lemma FilterNetwork_Trans_in_I_Ac : 
  forall {n} (M : @LTSNetwork n) (I : Ensemble nat) v s1 s2,
    NetTransSL (FilterNetwork M I) v s1 s2 ->
      forall i a, nth i (sync v) None = Some a -> i < n -> 
        Trans (nth i (Procs M) d_lts) a (nthS i s1) (nthS i s2).
Proof.
intros.
pose proof (FilterNetwork_Trans_Ac_in_I M I v s1 s2 H i (AcSome_prop v a i H0)).
destruct H.
simpl in H3.
specialize H3 with i.
rewrite H0 in H3.
rewrite Filter_nth in H3.
destruct (Ensemble2boollist_In n I i H1).
apply H4 in H2.
rewrite H2 in H3.
auto.
Qed.

(* A state that is filtered becomes a dummy state *)
Lemma FilterNetwork_States_not_in_I : 
  forall {n} (M : @LTSNetwork n) I s, States (SystemLTS (FilterNetwork M I)) s ->
    forall i, i < n -> ~I i -> nthS i s = d_state.
Proof.
intros n M I s HSs i Hin HNIi.
simpl in HSs.
unfold NetStates in HSs.
specialize HSs with i.
rewrite (FilterNetwork_Procs_i_not_in_I M I) in HSs ; try assumption.
simpl in HSs.
unfold SingletonS in HSs.
apply Singleton_inv in HSs.
rewrite HSs.
reflexivity.
exact Hin.
Qed.

Lemma FilterNetwork_States_in_I :
  forall {n} (M : @LTSNetwork n) I s, States (SystemLTS (FilterNetwork M I)) s ->
    forall i, i < n -> I i -> States (nth i (Procs M) d_lts) (nthS i s).
Proof.
intros.
simpl in H.
unfold NetStates in H.
specialize H with i.
rewrite <- (FilterNetwork_Procs_i_in_I M I) ; try assumption.
exact (H H0).
Qed.

Lemma FilterNetwork_States :
  forall {n} (M : @LTSNetwork n) I s, States (SystemLTS (FilterNetwork M I)) s <->
    (forall i, i < n -> 
         (I i -> States (nth i (Procs M) d_lts) (nthS i s))
      /\ (~I i -> nthS i s = d_state)).
Proof.
intuition.
apply (FilterNetwork_States_in_I M I) ; assumption.
apply (FilterNetwork_States_not_in_I M I) ; assumption.
simpl ; unfold NetStates ; intros i Hin.
specialize H with i.
destruct (EnsembleNat_dec I i) as [HI | HI] ;
  [ rewrite (FilterNetwork_Procs_i_in_I M I i Hin HI)
  | rewrite (FilterNetwork_Procs_i_not_in_I M I i Hin HI) ] ;
intuition.
rewrite H0.
apply In_singleton.
Qed.

Lemma DummyLTS_States :
  forall {n} (s : State n) i, States d_lts (nthS i s) -> nthS i s = d_state.
Proof.
simpl.
intros.
destruct H.
reflexivity.
Qed.

Lemma FilterNetwork_Init_not_in_I : 
  forall {n} (M : @LTSNetwork n) I s, Init (SystemLTS (FilterNetwork M I)) s ->
    forall i, i < n -> ~I i -> nthS i s = d_state.
Proof.
intros n M I s HSs i Hin HNIi.
simpl in HSs.
unfold NetInit in HSs.
specialize HSs with i.
rewrite (FilterNetwork_Procs_i_not_in_I M I) in HSs ; try assumption.
simpl in HSs.
unfold SingletonS in HSs.
apply Singleton_inv in HSs.
rewrite HSs.
reflexivity.
exact Hin.
Qed.

Lemma FilterNetwork_Init_in_I :
  forall {n} (M : @LTSNetwork n) I s, Init (SystemLTS (FilterNetwork M I)) s ->
    forall i, i < n -> I i -> Init (nth i (Procs M) d_lts) (nthS i s).
Proof.
intros.
simpl in H.
unfold NetInit in H.
specialize H with i.
rewrite <- (FilterNetwork_Procs_i_in_I M I) ; try assumption.
exact (H H0).
Qed.

Lemma FilterNetwork_Init :
  forall {n} (M : @LTSNetwork n) I s, Init (SystemLTS (FilterNetwork M I)) s <->
    (forall i, i < n -> 
         (I i -> Init (nth i (Procs M) d_lts) (nthS i s))
      /\ (~I i -> nthS i s = d_state)).
Proof.
intuition.
apply (FilterNetwork_Init_in_I M I) ; assumption.
apply (FilterNetwork_Init_not_in_I M I) ; assumption.
simpl ; unfold NetInit ; intros i Hin.
specialize H with i.
destruct (EnsembleNat_dec I i) as [HI | HI] ;
  [ rewrite (FilterNetwork_Procs_i_in_I M I i Hin HI) 
  | rewrite (FilterNetwork_Procs_i_not_in_I M I i Hin HI)] ; 
intuition.
rewrite H0.
apply In_singleton.
Qed.

Lemma DummyLTS_Init :
  forall {n} (s : State n) i, Init d_lts (nthS i s) -> nthS i s = d_state.
Proof.
simpl.
intros.
destruct H.
reflexivity.
Qed.

(* ==================================================================
     Composed state
   ================================================================== *)

Definition Cmp' {A}(bsp : option bool * (A * A)) :=
  match bsp with
  | (Some true, (_, a2))  => a2
  | (Some false, (a1, _)) => a1
  | (None, (a1, _))       => a1
  end.

Definition Cmp {n} (s : State n)(p : State n)(I : Ensemble nat) :=
  map Cmp' (combine (Ensemble2boollist n I) (combine s p)).

Lemma Cmp_length : forall {n} (s p : State n) I, length (Cmp s p I) = n.
Proof.
intros.
unfold Cmp.
rewrite map_length.
rewrite combine_length.
rewrite Ensemble2boollist_length.
rewrite combine_length.
rewrite lst_length.
rewrite lst_length.
assert (min n n = n) by intuition.
rewrite H.
exact H.
Qed.

(* s[p\I] := s where all i in I are replaced by p *)
Definition Cmps {n} (s : State n)(p : State n)(I : Ensemble nat) :=
  Bld_listn (Cmp s p I) (Cmp_length s p I).

Lemma Cmps_nth :
  forall {n} (s p : State n) I i,
    nth i (Cmps s p I) dummy = match (nth i (Ensemble2boollist n I) None) with
                               | Some true  => (nth i p dummy)
                               | Some false => (nth i s dummy)
                               | None       => (nth i s dummy) end.
Proof.
intros.
unfold Cmps, Cmp.
pose proof (map_nth (Cmp') (combine (Ensemble2boollist n I) (combine s p)) 
                           (None, (dummy, dummy)) i).
unfold Cmp', nthS in *.
simpl.
simpl in H.
rewrite H.
rewrite combine_nth.
rewrite combine_nth.
simpl.
destruct (nth i (Ensemble2boollist n I)) ; reflexivity.
rewrite (lst_length _ n s).
rewrite (lst_length _ n p).
reflexivity.
rewrite Ensemble2boollist_length.
rewrite combine_length.
rewrite lst_length.
rewrite lst_length.
assert (min n n = n) by intuition.
rewrite H0.
reflexivity.
Qed.

Lemma Cmps_In :
  forall {n} (s p : State n) (I : Ensemble nat) i, i < n ->
       (I i ->  nth i (Cmps s p I) dummy = nth i p dummy)
    /\ (~I i -> nth i (Cmps s p I) dummy = nth i s dummy).
Proof.
intros.
unfold nthS.
rewrite Cmps_nth.
destruct (Ensemble2boollist_In n I i H) as [HIn HNIn].
split ; intro ;
 [ apply HIn in H0 | apply HNIn in H0 ] ;
rewrite H0 ;
reflexivity.
Qed.

Lemma Cmps_i_in_I :
  forall {n} (s p : State n) (I : Ensemble nat) i, i < n -> I i ->  nth i (Cmps s p I) dummy = nth i p dummy.
Proof.
intros.
apply (Cmps_In s p I i H).
exact H0.
Qed.

Lemma Cmps_i_not_in_I :
  forall {n} (s p : State n) (I : Ensemble nat) i, i < n -> ~I i ->  nth i (Cmps s p I) dummy = nth i s dummy.
Proof.
intros.
apply (Cmps_In s p I i H).
exact H0.
Qed.

End Network.