(* Synchronisation uniformity conditions *)
Module SyncConditions.

Load "LTS.TransNetworkExtension".
Export TransNetworkExtension.
Load "Utils.OptionProps".
Import OptionProps.
Load "Utils.ClassicalExt".
Import ClassicalExt.

(* ------------------------------------------------------------------
     Dependency sets
   ------------------------------------------------------------------ *)

(* The direct dependency set *)
Definition DDep {n}(F : TransRule -> @LTS 1)(V : Ensemble (@SyncLaw n))(RS : @RuleSystem n)(i : nat)(j : nat) :=
  exists v, V v /\ OActs (F (nth i (Rules RS) d_tr)) (nth i (sync v) None)
                /\ OActs (F (nth j (Rules RS) d_tr)) (nth j (sync v) None).

(* The left and right direct dependency set *)
Definition DDepL {n}(RS : @RuleSystem n)(i : nat)(j : nat) := DDep L (SyncL RS) RS i j.
Definition DDepR {n}(RS : @RuleSystem n)(i : nat)(j : nat) := DDep R (Union SyncLaw (SyncL RS) (SyncR RS)) RS i j.

(* The dependency set *)
Definition Dep {n}(F : TransRule -> @LTS 1)(V : Ensemble (@SyncLaw n))(RS : @RuleSystem n) := 
  clos_trans nat (DDep F V RS).

(* The left and right dependency set *)
Definition DepL {n}(RS : @RuleSystem n) := Dep L (SyncL RS) RS.
Definition DepR {n}(RS : @RuleSystem n) := Dep R (Union SyncLaw (SyncL RS) (SyncR RS)) RS.

Lemma Dep_el_prop :
  forall {n} F V (RS : @RuleSystem n) i j , Dep F V RS i j -> j < n.
Proof.
intros.
induction H.
destruct H.
destruct H.
intuition.
destruct (OActs_In_to_Some _ _ H2) as [a [Hva Ha]].
exact (Acv_i_st_n _ _ (AcSome_prop _ _ _ Hva)).
assumption.
Qed.

Lemma Dep_sym :
  forall {n}(F : TransRule -> @LTS 1)(V : Ensemble (@SyncLaw n))(RS : @RuleSystem n) i j, 
    Dep F V RS i j <-> Dep F V RS j i.
Proof.
intros.
split ; intro H ; induction H as [x y | x y z] ;
  try first
  [ apply t_step ; unfold DDep in * ;
    destruct H as [v [Hv [HOAx HOAy]]] ;
    exists v ; intuition
  | apply (t_trans nat _ z y x) ; assumption].
Qed.

Lemma Dep_partitioning :
  forall {n}(F : TransRule -> @LTS 1)(V : Ensemble (@SyncLaw n))(RS : @RuleSystem n) i j,
    Dep F V RS i j -> (Dep F V RS i) = (Dep F V RS j).
Proof.
intros.
apply Extensionality_Ensembles.
unfold Same_set, Included, In.
split ; intros x H1 ;
    [ apply Dep_sym in H ; apply (t_trans nat _ j i x) ; assumption
    | apply (t_trans nat _ i j x) ; assumption ].
Qed.

Lemma Dep_partitioning_not_In :
  forall {n}(F : TransRule -> @LTS 1)(V : Ensemble (@SyncLaw n))(RS : @RuleSystem n) i j,
    ~Dep F V RS i j -> forall x, Dep F V RS j x -> ~Dep F V RS i x.
Proof.
intros.
unfold not ; intro H1 .
apply H.
rewrite (Dep_partitioning F V RS i x H1).
apply Dep_sym ; assumption.
Qed.

(* ------------------------------------------------------------------
     Synchronisation conditions
   ------------------------------------------------------------------ *)

Definition VCompleteness {n} (F : TransRule -> @LTS 1)(V : Ensemble (@SyncLaw n))(RS : @RuleSystem n) := 
  forall v, V v -> 
    (exists i, OActs (F (nth i (Rules RS) d_tr)) (nth i (sync v) None)) ->
      forall i, DActs (F (nth i (Rules RS) d_tr)) (nth i (sync v) None).

Definition cond1_universal_applicabilityL {n}{M}{RS}(TrM : @NetTransform n M RS) := 
  forall v i a, SyncL RS v ->
    let Li := L (nth i (Rules RS) d_tr) in
      As v a -> Acts Li a ->
        forall s s', Trans (nth i (Procs M) d_lts) a s s' ->
          exists q q', (mLs TrM i) q  s /\ (mLs TrM i) q' s' /\ Trans Li a q q'.

Definition cond2_universal_applicabilityR {n}{M}{RS}(TrM : @NetTransform n M RS) := 
  forall v i a, Union _ (SyncL RS) (SyncR RS) v ->
    let Ri := R (nth i (Rules RS) d_tr) in
      As v a -> Acts Ri a ->
        forall s s', Trans (nth i (Procs TrM) d_lts) a s s' ->
          exists q q', (mR _ (TrG TrM i)) q  s /\ (mR _ (TrG TrM i)) q' s' /\ Trans Ri a q q'.

Definition cond3_completenessL {n}(RS : @RuleSystem n) := VCompleteness L (SyncL RS) RS.

Definition cond4_completenessR {n}(RS : @RuleSystem n) := VCompleteness R (Union _ (SyncL RS) (SyncR RS)) RS.

Definition cond5_new_SyncLaw {n}(M : @LTSNetwork n)(RS : @RuleSystem n) :=
  forall v, SyncR RS v -> 
    forall i a, (nth i (sync v) None) = Some a -> 
         Acts (R (nth i (Rules RS) d_tr)) a /\ (~As v a -> ~Acts (nth i (Procs M) d_lts) a).

Section Synchronisation_Conditions.
Context {n} {M} {RS} (TrM : @NetTransform n M RS).

Lemma Ac_Fully_Matched_L :
  cond1_universal_applicabilityL TrM -> 
  cond3_completenessL RS ->
    forall v s s', NetTransSL M v s s' ->
      (exists i, exists q q', mLs TrM i q (nthS i s) /\ mLs TrM i q' (nthS i s') 
              /\ exists a, (nth i (sync v) None) = Some a /\ Trans (L (nth i (Rules RS) d_tr)) a q q')
        -> (forall i a, (nth i (sync v) None) = Some a -> 
              exists q q', mLs TrM i q (nthS i s) /\ mLs TrM i q' (nthS i s')
                         /\ Trans (L (nth i (Rules RS) d_tr)) a q q').
Proof.
intros Hcond1_universal_applicabilityL Hcond3_completenessL v s s' HT Hi i a Hva.
destruct HT as [Hv HT].
destruct Hi as [j [q [q' [Hmq [Hmq' [aj [Htj HTj]]]]]]].
destruct (classic (As v a)).
(* As M v a *)
  assert (i < n) as Hin by (apply (listn_nth_bound None (sync v) i) ; congruence).
  assert (j < n) as Hjn by (apply (listn_nth_bound None (sync v) j) ; congruence).
  specialize HT with i.
  destruct (trans_cond _ _ _ _ HTj) as [_ [_ Haj]].
  assert (exists x, OActs (L (nth x (Rules RS) d_tr)) (nth x (sync v) None)) 
    as HC by (exists j ; unfold OActs ; rewrite Htj ; exact Haj).
  rewrite <- (SyncMatch_cond TrM) in Hv.
  pose proof (Hcond3_completenessL v Hv HC i) as HOa.
  rewrite Hva in HOa, HT ; unfold OActs in HOa.
  destruct (Hcond1_universal_applicabilityL v i a Hv H HOa _ _ (HT Hin)) as [p [p' [Hmp [Hmp' HmT]]]].
  exists p, p'.
  intuition. 
(* ~As M v a *)
  pose proof (IndependentAction_prop M v a Hv H i j Hva (AcSome_prop _ _ _ Htj)).
  exists q, q'.
  rewrite H0 in Hva |- *.
  rewrite Htj in Hva.
  inversion Hva.
  rewrite <- H2.
  intuition.
Qed.

Lemma Ac_Fully_Matched_R :
  cond2_universal_applicabilityR TrM -> 
  cond4_completenessR RS ->
    forall v s s', NetTransSL TrM v s s' ->
      (exists i, exists q q', mR _ (TrG TrM i) q (nthS i s) /\ mR _ (TrG TrM i) q' (nthS i s') 
              /\ exists a, (nth i (sync v) None) = Some a /\ Trans (R (nth i (Rules RS) d_tr)) a q q')
        -> (forall i a, (nth i (sync v) None) = Some a -> 
              exists q q', mR _ (TrG TrM i) q (nthS i s) /\ mR _ (TrG TrM i) q' (nthS i s')
                         /\ Trans (R (nth i (Rules RS) d_tr)) a q q').
Proof.
intros Hcond2_universal_applicabilityR Hcond4_completenessR v s s' HT Hi i a Hva.
destruct HT as [Hv HT].
destruct Hi as [j [q [q' [Hmq [Hmq' [aj [Htj HTj]]]]]]].
destruct (classic (As v a)).
(* As M v a *)
  assert (i < n) as Hin by (apply (listn_nth_bound None (sync v) i) ; congruence).
  assert (j < n) as Hjn by (apply (listn_nth_bound None (sync v) j) ; congruence).
  specialize HT with i.
  destruct (trans_cond _ _ _ _ HTj) as [_ [_ Haj]].
  assert (exists x, OActs (R (nth x (Rules RS) d_tr)) (nth x (sync v) None)) 
    as HC by (exists j ; unfold OActs ; rewrite Htj ; exact Haj).
  rewrite (SyncN_cond TrM), <- (SyncMatch_cond TrM) in Hv.
  pose proof (Hcond4_completenessR v Hv HC i) as HOa.
  rewrite Hva in HOa, HT ; unfold OActs in HOa.
  destruct (Hcond2_universal_applicabilityR v i a Hv H HOa _ _ (HT Hin)) as [p [p' [Hmp [Hmp' HmT]]]].
  exists p, p'.
  intuition. 
(* ~As M v a *)
  pose proof (IndependentAction_prop TrM v a Hv H i j Hva (AcSome_prop _ _ _ Htj)).
  exists q, q'.
  rewrite H0 in Hva |- *.
  rewrite Htj in Hva.
  inversion Hva.
  rewrite <- H2.
  intuition.
Qed.

Lemma Ac_Mapped_prop_rewrite_helper : forall {A} {n} (P Q R : nat -> A -> Prop) (S T : Prop),
  (S /\ T /\ 
    exists p', forall j, j < n -> P j p' /\ Q j p' /\ R j p')
  ->
  (exists p', (S /\ (forall j, j < n -> P j p') /\ (forall j, j < n -> Q j p')) /\
               T /\ (forall j, j < n -> R j p')).
intuition.
destruct H2 as [p' HX].
exists p'.
intuition ; specialize HX with j ; intuition.
Qed.

(* If a sub-transition is matched, then the whole (System LTS) transition is matched *)
Lemma Ac_MappedL_prop :
  cond1_universal_applicabilityL TrM -> 
  cond3_completenessL RS ->
    forall v s s' sm i, let I := MatchLIds TrM s in
      NetTransSL M v s s' -> MapL TrM I sm s ->
        (exists q q', (mLs TrM i) q (nthS i s) /\ (mLs TrM i) q' (nthS i s') /\ 
           exists a : Act, nth i (sync v) None = Some a /\ Trans (L (nth i (Rules RS) d_tr)) a q q'
        ) -> exists s'm, MapL TrM I s'm s' /\ NetTransSL (LeftNetwork RS I) v sm s'm.
Proof.
intros Hcond1 Hcond3 v s s' sm i I HTM HMsm HT.
destruct HMsm as [HSs [HSsm HMsm]].
assert (forall i a, (nth i (sync v) None) = Some a -> exists q q', mLs TrM i q (nthS i s) 
           /\ mLs TrM i q' (nthS i s') /\ Trans (L (nth i (Rules RS) d_tr)) a q q')
  as HTi by ( apply (Ac_Fully_Matched_L Hcond1 Hcond3 v s s' HTM) ; exists i ; apply HT ).
destruct HT as [q [q' [Hmq [Hmq' [a [HAci HT]]]]]].
assert (i < n) as Hin by ( apply (listn_nth_bound None (sync v)) ; congruence ).
assert (Trans (SystemLTS M) (act v) s s') as HTX by (simpl ; exists v ; intuition).
destruct (trans_cond _ _ _ _ HTX) as [_ [HSs' _]] ; clear HTX.
destruct HTM as [Hv HTM].
apply Ac_Mapped_prop_rewrite_helper.
split ; try assumption.
split.
(* Show: SyncN (LeftNetwork RS I) v *)
simpl.
apply SyncF_all_i_not_in_I_None.
split.
intros j Hjn.
apply contrapositive_from_not.
intro Hva.
unfold not ; intro Htemp ; apply Htemp ; clear Htemp.
apply Option_NNone_Some in Hva.
destruct Hva as [aj Hvaj].
destruct (HTi j aj Hvaj) as [qj [_ [Hmqj _]]].
split ; try exact Hjn.
exists qj ; exact Hmqj.
simpl ; rewrite (SyncMatch_cond (M:=M)) ; assumption.
(* Show: exists s'm ... *)
set (P := fun j p' => 
  States (nth j (Procs (LeftNetwork RS I)) d_lts) (Bld_list1 p') /\
  (I j -> (mLs TrM j) (Bld_list1 p') (nthS j s')) /\
  match nth j (sync v) None with
  | Some x =>
      Trans (nth j (Procs (LeftNetwork RS I)) d_lts) x (nthS j sm) (Bld_list1 p')
  | None =>
      nth j sm dummy = p' /\
      States (nth j (Procs (LeftNetwork RS I)) d_lts) (nthS j sm) /\
      States (nth j (Procs (LeftNetwork RS I)) d_lts) (Bld_list1 p')
  end).
apply (listn_forall_exists _ n dummy P).
unfold P ; clear P
(* End rewrites *).
intros j Hjn.
case_eq (nth j (sync v) None) ; intros.
  (* v[j] = a0 *)
  destruct (HTi j a0 H) as [p [p' [Hmp [Hmp' HTL]]]].
  destruct (trans_cond _ _ _ _ HTL) as [HSp [HSp' _]].
  assert (MatchLIds TrM s j) as HMsj by
    ( split ; try exact Hjn ;
      exists p ; intuition ).
  assert (I j) as HIj by (assumption).
  rewrite (LeftNetwork_Procs_i_in_I RS I j Hjn HIj).
  exists (hd dummy p').
  rewrite list1_prop.
  intuition.
  assert (nthS j sm = p) as HRW by 
    ( apply (m_injective _ _ (mLs TrM j) _ _ (nthS j s)) ; right ; intuition ).
  rewrite HRW.
  exact HTL.
  (* v[j] = * *)
  specialize HMsm with j.
  specialize HTM with j.
  rewrite H in HTM.
  destruct (EnsembleNat_dec I j) as [HI | HI].
    (* j in I *)
    destruct HTM as [HSrw [HSsj HSs'j]] ; try exact Hjn.
    unfold nthS ; rewrite <- HSrw.
    exists (nth j sm dummy).
    rewrite (LeftNetwork_Procs_i_in_I RS I j Hjn HI).
    pose proof (LeftNetwork_States_in_I RS I sm HSsm j Hjn HI) as HSsmj.
    intuition.
    (* j not in I *)
    exists dummy.
    unfold LeftNetwork ; rewrite (FilterNetwork_Procs_i_not_in_I _ I j Hjn HI).
    pose proof (FilterNetwork_States_not_in_I _ I sm HSsm j Hjn HI) as HRW.
    simpl ; unfold SingletonS, d_state, nthS in *.
    apply list1_lst_eq in HRW.
    rewrite HRW.
    intuition.
Qed.

Lemma Ac_MappedR_prop :
  cond2_universal_applicabilityR TrM -> 
  cond4_completenessR RS ->
  forall v s s' sm i, let I := MatchRIds TrM s in
      NetTransSL TrM v s s' -> MapR TrM I sm s ->
        (exists q q', mR _ (TrG TrM i) q (nthS i s) /\ mR _ (TrG TrM i) q' (nthS i s') /\ 
           exists a : Act, nth i (sync v) None = Some a /\ Trans (R (nth i (Rules RS) d_tr)) a q q'
        ) -> exists s'm, MapR TrM I s'm s' /\ NetTransSL (RightNetwork RS I) v sm s'm.
Proof.
intros Hcond2 Hcond4 v s s' sm i I HTM HMsm HT.
destruct HMsm as [HSs [HSsm HMsm]].
assert (forall i a, (nth i (sync v) None) = Some a -> exists q q', mR _ (TrG TrM i) q (nthS i s) 
           /\ mR _ (TrG TrM i) q' (nthS i s') /\ Trans (R (nth i (Rules RS) d_tr)) a q q')
  as HTi by ( apply (Ac_Fully_Matched_R Hcond2 Hcond4 v s s' HTM) ; exists i ; apply HT ).
destruct HT as [q [q' [Hmq [Hmq' [a [HAci HT]]]]]].
assert (i < n) as Hin by ( apply (listn_nth_bound None (sync v)) ; congruence ).
assert (Trans (SystemLTS TrM) (act v) s s') as HTX by (simpl ; exists v ; intuition).
destruct (trans_cond _ _ _ _ HTX) as [_ [HSs' _]] ; clear HTX.
destruct HTM as [Hv HTM].
apply Ac_Mapped_prop_rewrite_helper.
split ; try assumption.
split.
(* Show: SyncN (RightNetwork RS I) v *)
simpl.
apply SyncF_all_i_not_in_I_None.
split.
intros j Hjn.
apply contrapositive_from_not.
intro Hva.
unfold not ; intro Htemp ; apply Htemp ; clear Htemp.
apply Option_NNone_Some in Hva.
destruct Hva as [aj Hvaj].
destruct (HTi j aj Hvaj) as [qj [_ [Hmqj _]]].
split ; try exact Hjn.
exists qj ; exact Hmqj.
simpl ; rewrite (SyncMatch_cond (M:=M)) ; try rewrite <- (SyncN_cond TrM) ; assumption.
(* Show: exists s'm ... *)
set (P := fun j p' => 
  States (nth j (Procs (RightNetwork RS I)) d_lts) (Bld_list1 p') /\
  (I j -> mR _ (TrG TrM j) (Bld_list1 p') (nthS j s')) /\
  match nth j (sync v) None with
  | Some x =>
      Trans (nth j (Procs (RightNetwork RS I)) d_lts) x (nthS j sm) (Bld_list1 p')
  | None =>
      nth j sm dummy = p' /\
      States (nth j (Procs (RightNetwork RS I)) d_lts) (nthS j sm) /\
      States (nth j (Procs (RightNetwork RS I)) d_lts) (Bld_list1 p')
  end).
apply (listn_forall_exists _ n dummy P).
unfold P ; clear P
(* End rewrites *).
intros j Hjn.
case_eq (nth j (sync v) None) ; intros.
  (* v[j] = a0 *)
  destruct (HTi j a0 H) as [p [p' [Hmp [Hmp' HTL]]]].
  destruct (trans_cond _ _ _ _ HTL) as [HSp [HSp' _]].
  assert (MatchRIds TrM s j) as HMsj by
    ( split ; try exact Hjn ;
      exists p ; intuition ).
  assert (I j) as HIj by (assumption).
  rewrite (RightNetwork_Procs_i_in_I RS I j Hjn HIj).
  exists (hd dummy p').
  rewrite list1_prop.
  intuition.
  assert (nthS j sm = p) as HRW.
    apply (m_injective _ _ (mR _ (TrG TrM j)) _ _ (nthS j s)).
    specialize HMsm with j ; right ; intuition
  (* End assert *).
  rewrite HRW.
  exact HTL.
  (* v[j] = * *)
  specialize HMsm with j.
  specialize HTM with j.
  rewrite H in HTM.
  destruct (EnsembleNat_dec I j) as [HI | HI].
    (* j in I *)
    destruct HTM as [HSrw [HSsj HSs'j]] ; try exact Hjn.
    unfold nthS ; rewrite <- HSrw.
    exists (nth j sm dummy).
    rewrite (RightNetwork_Procs_i_in_I RS I j Hjn HI).
    pose proof (RightNetwork_States_in_I RS I sm HSsm j Hjn HI) as HSsmj.
    intuition.
    (* j not in I *)
    exists dummy.
    unfold RightNetwork ; rewrite (FilterNetwork_Procs_i_not_in_I _ I j Hjn HI).
    pose proof (FilterNetwork_States_not_in_I _ I sm HSsm j Hjn HI) as HRW.
    simpl ; unfold SingletonS, d_state, nthS in *.
    apply list1_lst_eq in HRW.
    rewrite HRW.
    intuition.
Qed.

Lemma NAs_universal_applicabilityR :
  cond5_new_SyncLaw M RS ->
    forall v i a, let Ri := R (nth i (Rules RS) d_tr) in
      ~As v a -> SyncR RS v -> nth i (sync v) None = Some a -> Acts Ri a ->
        forall s s', Trans (nth i (Procs TrM) d_lts) a s s' ->
          exists q q', (mR _ (TrG TrM i)) q  s /\ (mR _ (TrG TrM i)) q' s' /\ Trans Ri a q q'.
Proof.
intros Hcond5_new_SyncLaw v i a Ri HAs Hv Hva HRa s s' HTTrM.
unfold Ri in * ; clear Ri.
destruct (Hcond5_new_SyncLaw v Hv i a Hva) as [_ HNGa'].
pose proof (HNGa' HAs) as HNGa.
pose proof (listn_nth_bound _ _ _ (Option_Some_NNone _ _ Hva)) as Hin.
rewrite (Procs_cond TrM i Hin) in HTTrM.
apply (TT_cond _ (TrG TrM i) a s s') in HTTrM.
destruct HTTrM as [[HT HNM] | HM].
  (* Trans was not added *)
  pose proof (trans_cond _ _ _ _ HT).
  intuition.
  (* Trans was added *)
  exact HM.
Qed.

Lemma unchanged_transitions_v_in_SynN_M :
  cond2_universal_applicabilityR TrM ->
  cond5_new_SyncLaw M RS ->
    forall v p p', NetTransSL TrM v p p' ->
      (exists i, Ac v i /\
         ~(exists q q', (mR _ (TrG TrM i)) q (nthS i p) /\ (mR _ (TrG TrM i)) q' (nthS i p') 
             /\ exists ai, nth i (sync v) None = Some ai /\ Trans (R (nth i (Rules RS) d_tr)) ai q q'))
        -> SyncN M v.
Proof.
intros Hcond2_universal_applicabilityR Hcond5_new_SyncLaw v p p' HT HM.
destruct HT as [HTrMv HT].
pose proof HTrMv as Hv.
rewrite SyncN_cond in Hv.
destruct Hv as [v Hv | v Hv]; try exact Hv.
destruct HM as [i [HAci HM]].
pose proof (Acv_i_st_n _ _ HAci) as Hin.
pose proof HAci as [a Hva].
pose proof (HT i Hin) as HTi.
rewrite Hva in HTi.
destruct (Hcond5_new_SyncLaw v Hv i a Hva) as [HRa HCNAs].
rewrite (SyncN_cond TrM), <- (SyncMatch_cond TrM) in HTrMv.
exfalso.
destruct (classic (As v a)) as [HAs | HAs].
(* As v a *)
apply HM.
destruct (Hcond2_universal_applicabilityR v i a HTrMv HAs HRa _ _ HTi) as [q [q' Hm]].
exists q, q' ; intuition.
exists a ; intuition.
(* ~As v a *)
apply HM.
destruct (NAs_universal_applicabilityR Hcond5_new_SyncLaw  v i a HAs Hv Hva HRa _ _ HTi) as [q [q' HTM]].
exists q, q' ; intuition.
exists a ; intuition.
Qed.

End Synchronisation_Conditions.

End SyncConditions.