(* Extended network and related notions *)

Module TransNetworkExtension.

Load "LTS.TransformationNetwork".
Export TransformationNetwork.
Load "Utils.ClassicalExt".
Import ClassicalExt.

(* ==================================================================
     System of kappa extended transformation rules
   ================================================================== *)

Definition boolState2Act (k_fun : KFun)(p : option bool * state_type) : option Act :=
  match p with
  | (Some true, s) => Some (k_fun (Bld_list1 s))
  | otherwise      => None
  end.

Lemma KSyncVec_length : 
  forall {n} k_fun I (s : State n),
    length (map (boolState2Act k_fun) (combine (Ensemble2boollist n I) s)) = n.
Proof.
intros.
rewrite map_length.
rewrite combine_length.
rewrite Ensemble2boollist_length.
rewrite lst_length.
intuition.
Qed.

(* The kappa-synchronisation vector *)
Definition KSyncVec {n} (k_fun : KFun)(I : Ensemble nat)(s : State n) :=
  Build_listn (option Act) n
    (map (boolState2Act k_fun) (combine (Ensemble2boollist n I) s))
    (KSyncVec_length k_fun I s).

Lemma KSyncVec_prop :
  forall {n} k_fun (I : Ensemble nat) (s : State n),
       (forall i, i < n -> forall a, nth i (KSyncVec k_fun I s) None = Some a -> I i /\ a = k_fun (nthS i s))
    /\ (forall i, i < n -> nth i (KSyncVec k_fun I s) None = None -> ~I i).
Proof.
unfold KSyncVec.
unfold boolState2Act.
intros.
assert (Datatypes.length (Ensemble2boollist n I) = Datatypes.length s) as Hlength.
  rewrite (Ensemble2boollist_length n I).
  rewrite (lst_length _ _ s).
  reflexivity (* End assert *).
unfold nthS.
simpl.
remember (combine (Ensemble2boollist n I) s) as I'.
pose (map_nth (boolState2Act k_fun) I' (None, dummy)) as RWmap.
unfold boolState2Act in RWmap.
simpl in RWmap.
split ; intros i Hin ; [intros a HSa | intros HK] ;
rewrite (RWmap i) in * ;
clear RWmap ;
rewrite HeqI' in * ;
rewrite (combine_nth (Ensemble2boollist n I) s i None dummy Hlength) in * ;
destruct (Ensemble2boollist_In n I i Hin) as [[_ HIn] [_ HNIn]] ;
destruct (EnsembleNat_dec I i) as [HI | HI] ; try rewrite  (HIn HI) in * ; try rewrite (HNIn HI) in * ;
simpl in * ; try inversion HSa ; intuition.
congruence.
Qed.

Lemma KSyncVec_applicable :
  forall {n} k_fun I J v (s : State n), sync v = KSyncVec k_fun (Intersection state_type I J) s ->
    SyncLaw_applicable I v.
Proof.
intros n k_fun I J v s HK.
apply (SyncLaw_applicable_all_i_not_in_I_None I v).
rewrite HK.
intros i Hin HI.
destruct (Ensemble2boollist_In n (Intersection _ I J) i Hin) as [_ [_ HIJNIn]].
assert (~Intersection _ I J i) as HIJi by
  (unfold not ; intro HIJi; apply HI; destruct HIJi ; assumption).
unfold KSyncVec in *.
simpl in *.
pose proof (map_nth (boolState2Act k_fun) (combine (Ensemble2boollist n (Intersection _ I J)) s) (None, (nth i s dummy)) i) as HGR.
simpl in HGR.
rewrite HGR.
rewrite combine_nth.
simpl.
rewrite (HIJNIn HIJi).
reflexivity.
rewrite Ensemble2boollist_length, lst_length ; reflexivity.
Qed.

(* The kappa synchronisation law *)
Definition SyncK {n} (RS : @RuleSystem n)(k_fun : KFun)(k_fun2 : listn (option Act) n -> Act)(v : SyncLaw) :=
     (act v) = k_fun2 (sync v)
  /\ (exists I s i, i < n /\ I i
                 /\ Init (L (nth i (Rules RS) d_tr)) (nthS i s)
                 /\ sync v = KSyncVec k_fun I s).

Lemma exists_v_in_SyncK :
  forall {n} RS k_fun k_fun2 I (s : State n), 
    (exists i, i < n /\ I i /\ Init (L (nth i (Rules RS) d_tr)) (nthS i s)) ->
      exists v, SyncK RS k_fun k_fun2 v /\ sync v = KSyncVec k_fun I s.
Proof.
intros.
destruct H as [i [Hin [HI HL]]].
assert (exists i, nth i (KSyncVec k_fun I s) None <> None) as HC.
  exists i.
  destruct (KSyncVec_prop k_fun I s) as [_ HK].
  apply (contrapositive_to_not _ _ (HK i Hin)).
  unfold not ; intros ; contradiction.
exists (Build_SyncLaw n (KSyncVec k_fun I s) (k_fun2 (KSyncVec k_fun I s)) HC).
unfold SyncK.
intuition.
exists I, s, i.
intuition.
Qed.

(* A function that produces a (result) kappa-synchronisation action from a kappa synchronisation vector *)
Record KFun2 {n}(RS : RuleSystem)(k_fun : KFun) := {
  kfun2 :> listn (option Act) n -> Act;

  kfun2_not_tau : forall u, kfun2 u <> tau;

  (* If v is a kappa-synclaw and (act v') = (act v), then we must have v' = v 
     (thus v' is also a kappa-synclaw), i.e. the action of a kappa-synclaw is unique. *)
  SyncK_cond : forall v v', SyncK RS k_fun kfun2 v /\ act v' = act v -> v' = v;

  (* The action of a kappa-synclaw is fresh. *)
  Kact_cond  : forall I u, let SysL := SystemLTS (LeftNetwork RS I) in
                           let SysR := SystemLTS (RightNetwork RS I) in
                 ~Acts SysL (kfun2 u) /\ ~Acts SysR (kfun2 u);

  (* Kappa-actions may not occur in the synchronisation vector of the original
     (un-extended) rule system *)
  SyncVec_cond : forall v i a, 
                   (SyncL RS v \/ (SyncR RS v)) /\ (nth i (sync v) None = Some a) -> 
                     NoKAct (nth i (Rules RS) d_tr) k_fun a
}.

(* ------------------------------------------------------------------
     kappa extended rule system
   ------------------------------------------------------------------ *)

(* System of kappa-extended transformation rules *)
Record RuleSystemK {n}{RS : @RuleSystem n}{k_fun : KFun}(k_fun2 : KFun2 RS k_fun) := {
  rsk :> RuleSystem;

  rki (i : nat) : TransRuleK (nth i (Rules RS) d_tr) k_fun;

  r_cond  : forall i, i < n -> nth i (Rules rsk) d_tr = rki i;

  SyncL_cond : forall v, (SyncL rsk) v <-> (SyncL RS) v \/ SyncK RS k_fun k_fun2 v;
  SyncR_cond : forall v, (SyncR rsk) v <-> (SyncR RS) v \/ SyncK RS k_fun k_fun2 v
}.

Lemma S_cond :
  forall {n} {RS} {k_fun} {k_fun2} (RSk : @RuleSystemK n RS k_fun k_fun2) I,
    let SysL (R : RuleSystem) := SystemLTS (LeftNetwork R I) in
    let SysR (R : RuleSystem) := SystemLTS (RightNetwork R I) in
         States (SysL RSk) = States (SysL RS)
      /\ States (SysR RSk) = States (SysR RS).
Proof.
intros.
unfold SysL, SysR.
split ;
apply Extensionality_Ensembles ;
unfold Same_set, Included, In ;
intuition ; first 
  [ simple apply (LeftNetwork_States_lift _ I x) ; intros ;
    simple apply (LeftNetwork_States_simpl _ I x i) in H
  | simple apply (RightNetwork_States_lift _ I x) ; intros ;
    simple apply (RightNetwork_States_simpl _ I x i) in H] ;
try exact H0 ;
simpl in * ; 
try rewrite r_cond in H ; try rewrite r_cond ; auto ;
remember (rki k_fun2 RSk i) as rk ;
destruct (s_cond (nth i (Rules RS) d_tr) k_fun rk) as [HL HR] ;
try rewrite HL in * ;
try rewrite HR in * ;
split ; intuition.
Qed.

(* The initial states of the kappa-extended system and the original system are identical *)
Lemma I_cond : 
  forall {n} {RS} {k_fun} {k_fun2} (RSk : @RuleSystemK n RS k_fun k_fun2) I,
    let SysL (R : RuleSystem) := SystemLTS (LeftNetwork R I) in
    let SysR (R : RuleSystem) := SystemLTS (RightNetwork R I) in
         Init (SysL RSk) = Init (SysL RS)
      /\ Init (SysR RSk) = Init (SysR RS).
Proof.
intros.
unfold SysL, SysR.
split ;
apply Extensionality_Ensembles ;
unfold Same_set, Included, In ;
intuition ; first 
  [ simple apply (LeftNetwork_Init_lift _ I x) ; intros ;
    simple apply (LeftNetwork_Init_simpl _ I x i) in H
  | simple apply (RightNetwork_Init_lift _ I x) ; intros ;
    simple apply (RightNetwork_Init_simpl _ I x i) in H] ;
try exact H0 ;
simpl in * ; 
try rewrite r_cond in H ; try rewrite r_cond ; auto ;
remember (rki k_fun2 RSk i) as rk ;
destruct (i_cond (nth i (Rules RS) d_tr) k_fun rk) as [HL HR] ;
try rewrite HL in * ;
try rewrite HR in * ;
split ; intuition.
Qed.

(* A transition in RS is also present in kappa-extended RS *)
Lemma Trans_RS_then_Trans_RSk :
  forall {n} {RS} {k_fun} {k_fun2} (RSk : @RuleSystemK n RS k_fun k_fun2) I v s s',
    let NetL (R : RuleSystem) := LeftNetwork R I in
    let NetR (R : RuleSystem) := RightNetwork R I in
       (NetTransSL (NetL RS) v s s' -> NetTransSL (NetL RSk) v s s')
    /\ (NetTransSL (NetR RS) v s s' -> NetTransSL (NetR RSk) v s s').
Proof.
intros.
unfold NetL, NetR.
pose proof (SyncL_cond k_fun2 RSk v) as [_ H].
pose proof (SyncR_cond k_fun2 RSk v) as [_ H1].
split ; intros HT ;
pose proof (FilterNetwork_Trans_Ac_in_I _ I v s s' HT) as HAcI ;
destruct HT as [Hv HT] ; split ;
try (destruct Hv as [Hv HFv] ; split) ; intuition ;
  first [ destruct Hv as [Hv | Hv] ; [ apply Union_introl | apply Union_intror ]
  | specialize HT with i ;
    destruct (s_cond (nth i (Rules RS) d_tr) k_fun (rki k_fun2 RSk i)) as [HL HR] ;
    destruct (EnsembleNat_dec I i) as [HI | HI] ;
    [ (* I i *)
      try rewrite (LeftNetwork_Procs_i_in_I RSk I i H1 HI) ;
      try rewrite (LeftNetwork_Procs_i_in_I RS I i H1 HI) in HT ;
      try rewrite (RightNetwork_Procs_i_in_I RSk I i H1 HI) ;
      try rewrite (RightNetwork_Procs_i_in_I RS I i H1 HI) in HT ;
      rewrite r_cond ; try assumption ;
      try rewrite HL ; try rewrite HR ;
      destruct (nth i (sync v) None) as [a Hva | Hv] ;
      try apply (t_cond (nth i (Rules RS) d_tr) k_fun _ a (nthS i s) (nthS i s'))
    | (* ~I i *)
      unfold LeftNetwork, RightNetwork in * ;
      rewrite (FilterNetwork_Procs_i_not_in_I _ I i H1 HI) ;
      rewrite (FilterNetwork_Procs_i_not_in_I _ I i H1 HI) in HT ; specialize HAcI with i ; 
      rewrite (NotAc_to_SyncLawNone v i (contrapositive_to_not _ _ HAcI HI)) in HT |- * 
    ]
  ] ; intuition.
Qed.

(* A transition in kappa-extened RS is also present in RS if it is not a kappa-transition *)
Lemma Trans_RSk_then_Trans_RS :
  forall {n} {RS} {k_fun} {k_fun2} (RSk : @RuleSystemK n RS k_fun k_fun2) I v s s',
    let NetL (R : RuleSystem) := LeftNetwork R I in
    let NetR (R : RuleSystem) := RightNetwork R I in
       (SyncN (NetL RS) v /\ NetTransSL (NetL RSk) v s s' -> NetTransSL (NetL RS) v s s')
    /\ (SyncN (NetR RS) v /\ NetTransSL (NetR RSk) v s s' -> NetTransSL (NetR RS) v s s').
Proof.
simpl.
unfold NetTransSL.
simpl.
intuition ; specialize H2 with i ;
rewrite Filter_nth in H2 ; rewrite Filter_nth ;
simpl in H2 ; simpl ;
first [ rewrite L_map_nth in H2 ; rewrite L_map_nth
      | rewrite R_map_nth in H2 ; rewrite R_map_nth ] ;
rewrite r_cond in H2 ; auto ;
remember (rki k_fun2 RSk i) as rk ;
destruct (s_cond (nth i (Rules RS) d_tr) k_fun rk) as [HL HR] ;
destruct (Ensemble2boollist_In n I i H1) as [[_ HIn] [_ HNIn]] ;
(destruct (EnsembleNat_dec I i) as [HI | HI] ; [rewrite  (HIn HI) in * | rewrite (HNIn HI) in *]) ;
intuition ;
case_eq (nth i (sync v) None) ; first [intros a HS | intro HS] ; rewrite HS in H3 ;
first
  [ rewrite <- HL ; auto
  | rewrite <- HR ; auto
  | intuition ;
    destruct (Trans_rk_then_Trans_r2 (nth i (Rules RS) d_tr) k_fun 
              rk a (nthS i s) (nthS i s')) as [HKL HKR] ;
    try apply HKL ; try apply HKR ;
    pose proof (SyncVec_cond _ _ k_fun2 v i a) ;
    try destruct H0 ; try destruct H0 ;
    intuition
  ].
Qed.

(* A transition in kappa-extened RS is also present in RS *)
Lemma TauPath_RSk_then_TauPath_RS :
  forall {n} {RS} {k_fun} {k_fun2} (RSk : @RuleSystemK n RS k_fun k_fun2) I s s',
    let SysL (R : RuleSystem) := SystemLTS (LeftNetwork R I) in
    let SysR (R : RuleSystem) := SystemLTS (RightNetwork R I) in 
         (TauPath (SysL RSk) s s' -> TauPath (SysL RS) s s')
      /\ (TauPath (SysR RSk) s s' -> TauPath (SysR RS) s s').
Proof.
intros.
intuition.
induction H.
destruct H as [v HT].
(* Case: Trans G x y -> Trans G x y *)
apply rt_step.
exists v ; intuition.
apply (Trans_RSk_then_Trans_RS RSk).
split ; try assumption.
unfold NetTransSL in H0.
simpl in *.
unfold SyncF in *.
destruct (SyncL_cond k_fun2 RSk v) as [HS _].
unfold SyncK in HS.
intuition.
exfalso ; apply (kfun2_not_tau RS k_fun k_fun2 (sync v)).
rewrite <- H1.
exact H.
(* Case: TauPath G s s *)
apply rt_refl.
(* Case: TauPath G s sh /\ TauPath G sh s' -> TauPath G s s' *)
apply (rt_trans (State n) (Trans (SysL RS) tau) x y) ; assumption.
induction H.
(* Case: Trans G x y -> Trans G x y *)
apply rt_step.
destruct H as [v HT].
exists v ; intuition.
apply (Trans_RSk_then_Trans_RS RSk).
split ; try assumption.
unfold NetTransSL in H0.
simpl in *.
destruct H0.
destruct H0.
split ; try assumption.
destruct H0 ; 
  [ apply Union_introl ; destruct (SyncL_cond k_fun2 RSk x0) as [HS _]
  | apply Union_intror ; destruct (SyncR_cond k_fun2 RSk x0) as [HS _] ] ;
unfold SyncK in HS ; intuition ;
exfalso ; apply (kfun2_not_tau RS k_fun k_fun2 (sync x0)) ;
rewrite <- H3 ; exact H.
(* Case: TauPath G s s *)
apply rt_refl.
(* Case: TauPath G s sh /\ TauPath G sh s' -> TauPath G s s' *)
apply (rt_trans (State n) (Trans (SysR RS) tau) x y) ; assumption.
Qed.

(* kappa-synchronisation laws are uniquely present on glue-(sub-)states *)
Lemma K_cond :
  forall {n} {RS} {k_fun} {k_fun2} (RSk : @RuleSystemK n RS k_fun k_fun2) (I J : Ensemble nat) s,
    let SysL (R : RuleSystem) := SystemLTS (LeftNetwork R I) in 
    let SysR (R : RuleSystem) := SystemLTS (RightNetwork R I) in
    let r i := (nth i (Rules RS) d_tr) in
      (forall i, i < n -> I i -> J i -> Init (L (r i)) (nthS i s)) -> 
      (exists i, i < n /\ I i /\ J i) ->
         exists k, (States (SysL RS) s -> Trans (SysL RSk) k s s) 
                /\ (States (SysR RS) s -> Trans (SysR RSk) k s s)
                /\ k <> tau
                /\ forall s1 s2, Trans (SysL RSk) k s1 s2 \/ Trans (SysR RSk) k s1 s2
                -> forall i, i < n -> I i -> J i -> 
                        nth i s1 dummy = nth i s dummy 
                     /\ nth i s2 dummy = nth i s dummy.
Proof.
intros.
pose proof H as HILs ; clear H.
destruct H0 as [i [Hin [HIi HJi]]].
assert (forall i : nat, i < n -> I i -> J i -> Init (R (r i)) (nthS i s)) as HIRs by
  ( intros ; destruct (glue_cond (r i0)) as [HG _] ;
    rewrite <- HG ; exact (HILs i0 H H0 H1) ).
destruct (S_cond RSk I) as [HLrw HRrw].
simpl in HLrw, HRrw |- *.
rewrite <- HLrw, <- HRrw.
clear SysL SysR HLrw HRrw.

pose proof (HILs i) as HIsi.
destruct (exists_v_in_SyncK RS k_fun k_fun2 (Intersection nat I J) s) as [v [Hv1 Hv2]].
exists i.
intuition.
split ; assumption.
clear HIi HJi Hin HIsi i.
exists (act v).
destruct (KSyncVec_prop k_fun (Intersection nat I J) s) as [HKSome HKNone].
rewrite <- Hv2 in HKSome, HKNone.
(* Case: Trans (SysL RSk) k s s /\ Trans (SysR RSk) k s s *)
split ; try split ;
  try (
    unfold NetTrans ;
    unfold NetTransSL ;
    simpl ;
    exists v ;
    intuition ;
    first 
    [ (* Sync RSk v *)
      split ; [ try left ; apply SyncL_cond ; right ; exact Hv1 
              | apply (KSyncVec_applicable k_fun I J v s) ; exact Hv2 ]
    | (* Trans (Sys RSk) k s s *)
      pose proof H0 as Hin ; clear H0 ;
      specialize HILs with i ;
      specialize HIRs with i ;
      unfold NetStates in H ;
      specialize H with i ;
      simpl in H ;
      rewrite Filter_nth in H ; rewrite Filter_nth ;
      simpl in H ; simpl ;
      first [ rewrite R_map_nth in H ; rewrite R_map_nth
            | rewrite L_map_nth in H ; rewrite L_map_nth ] ;
      rewrite r_cond in H ; auto ; rewrite r_cond ; auto ; 
      remember (rki k_fun2 RSk i) as rk ;
      case_eq (nth i (sync v) None) ; [intros a Ha | intro Ha] ;
      [ (* Case: v[i] = a *)
        destruct (HKSome i Hin a Ha) as [HIJi Hka] ;
        destruct HIJi as [i HIi HJi] ; unfold In in HIi, HJi ;
        destruct (t_cond (nth i (Rules RS) d_tr) k_fun rk a (nthS i s) (nthS i s)) as [HTL HTR] ;
        destruct (Ensemble2boollist_In n I i) as [[_ HIn] _] ; try exact Hin ;
        rewrite (HIn HIi) in * ; clear HIn ;
        apply HTL || apply HTR ; right ; intuition
      | (* Case: v[i] = None *)
        destruct (nth i (Ensemble2boollist n I) None) ;
        intuition ]
  ] ).
(* Case: k <> tau *)
split.
unfold SyncK in Hv1.
destruct Hv1 as [Hvtau _].
rewrite Hvtau.
apply (kfun2_not_tau RS k_fun k_fun2 (sync v)).
(* Case: the k-action only occurs on state s *)
intros.
pose proof (trans_cond (SystemLTS (LeftNetwork RSk I)) (act v) s1 s2) as HLS.
pose proof (trans_cond (SystemLTS (RightNetwork RSk I)) (act v) s1 s2) as HRS.
simpl in HLS, HRS.
destruct H ;
  [ destruct (HLS H) as [HSs1 [HSs2 _]]
  | destruct (HRS H) as [HSs1 [HSs2 _]] ] ;
clear HLS HRS ;
destruct H as [v' [Hv'a [_ Hv'T]]] ;
simpl in Hv'T ;
pose proof (SyncK_cond RS k_fun k_fun2 v v') ;
intuition ;
specialize Hv'T with i ;
rewrite H3 in Hv'T ;
case_eq (nth i (sync v) None) ; intros ; rewrite H in Hv'T ;
  first 
  [ (* Case: v[i] = a *)
    specialize HKSome with i a ;
    apply HKSome in H ; auto ; clear H3 Hv'a ;
    destruct H as [_ aRW] ;
    rewrite Filter_nth in Hv'T ;
    simpl in Hv'T ;
    first [ rewrite L_map_nth in Hv'T
          | rewrite R_map_nth in Hv'T ] ;
    rewrite r_cond in Hv'T ; auto ; remember (rki k_fun2 RSk i) as rk ;
    destruct (t_cond (nth i (Rules RS) d_tr) k_fun rk a (nthS i s1) (nthS i s2)) as [HTL HTR] ;
    destruct (Ensemble2boollist_In n I i) as [HIn _] ; auto ;
    apply HIn in H1 ;
    rewrite H1 in Hv'T ;
    try apply HTL in Hv'T ; try apply HTR in Hv'T ; auto ;
    clear HTL HTR ;
    pose proof (kact_cond (nth i (Rules RS) d_tr) k_fun rk (nthS i s)) ;
    rewrite aRW in Hv'T ;
    apply list1_lst_eq ;
    first [ apply (kfun_injective k_fun (nthS i s1) (nthS i s)) 
          | apply (kfun_injective k_fun (nthS i s2) (nthS i s)) ] ;
    pose proof (trans_cond (L (nth i (Rules RS) d_tr)) 
                  (k_fun (nthS i s)) (nthS i s1) (nthS i s2)) ;
    pose proof (trans_cond (R (nth i (Rules RS) d_tr)) 
                  (k_fun (nthS i s)) (nthS i s1) (nthS i s2)) ;
    intuition ; 
    try rewrite <- H8 ;
    intuition
  | (* Case: v[i] = * *)
    exfalso ;
    apply (HKNone i H0 H) ;
    split ; assumption ].
Qed.

Lemma KI_cond :
  forall {n} {RS} {k_fun} {k_fun2} (RSk : @RuleSystemK n RS k_fun k_fun2) I s,
    let SysL (R : RuleSystem) := SystemLTS (LeftNetwork R I) in 
    let SysR (R : RuleSystem) := SystemLTS (RightNetwork R I) in 
      Init (SysL RS) s -> (exists i, i < n /\ I i) ->
           exists k, Trans (SysL RSk) k s s /\ Trans (SysR RSk) k s s /\ k <> tau
        /\ forall s1 s2, Trans (SysL RSk) k s1 s2 \/ Trans (SysR RSk) k s1 s2
             -> s1 = s /\ s2 = s.
Proof.
intros.
destruct H0 as [i [Hin Hi]].
pose proof (init_cond (SysL RS) s H) as HLSs.
unfold SysL in HLSs, H.
pose proof H as HRSs.
destruct (Glue_cond RS I) as [HG _].
rewrite HG in HRSs ; clear HG.
apply (init_cond (SystemLTS (RightNetwork RS I)) s) in HRSs.
unfold SysL, SysR ; clear SysL SysR.
pose proof (K_cond RSk I I s).
destruct H0 as [k Hk].
intros ; apply (LeftNetwork_Init_simpl RS I s i0) in H ; intuition.
exists i ; intuition.
exists k.
intuition ;
destruct (trans_cond _ _ _ _ H5) as [HSs1 [HSs2 _]] ;
specialize H4 with s1 s2 i ;
intuition ;
apply (lstn_eq _ _ dummy) ;
intros i0 Hi0n ;
destruct (EnsembleNat_dec I i0) as [HI | HI] ;
  solve [
    (* I i0 *)
    apply H4 ; assumption
  | (* ~I i0 *)
    apply list1_lst_eq ;
    pose proof (FilterNetwork_States_not_in_I _ _ _ HSs1 i0 Hi0n HI) as Hs1 ;
    pose proof (FilterNetwork_States_not_in_I _ _ _ HSs2 i0 Hi0n HI) as Hs2 ;
    pose proof (FilterNetwork_States_not_in_I _ _ _ HLSs i0 Hi0n HI) as Hs ;
    unfold nthS in Hs1, Hs2, Hs ;
    rewrite Hs ; try rewrite Hs1 ; try rewrite Hs2 ;
    reflexivity
  ].
Qed.

(* If a vector state s, of which some sub-states are glue-states, is related to p, then there exists 
   a state ph with the same glue-sub-states as s such that p ==> ph and ph is related to s as well *)
Lemma TauPath_glue_state_L_prop : 
  forall {n} {RS} {k_fun} {k_fun2} (RSk : @RuleSystemK n RS k_fun k_fun2) (I J : Ensemble nat),
    let SysL (R : RuleSystem) := SystemLTS (LeftNetwork R I) in 
    let SysR (R : RuleSystem) := SystemLTS (RightNetwork R I) in
      forall (B : BB (SysL RSk) (SysR RSk)) s p,
           B s p -> (forall i, i < n -> I i -> J i -> Init (L (nth i (Rules RS) d_tr)) (nthS i s))
            -> exists ph, TauPath (SysR RS) p ph /\ B s ph 
                         /\ (forall i, i < n -> I i -> J i -> nth i ph dummy = nth i s dummy).
Proof.
intros.
destruct (B_states _ _ _ _ _ H).
pose proof (B_branching (SysL RSk) (SysR RSk) B s p H).
unfold SysL, SysR in * ; clear SysL SysR.
destruct (S_cond RSk I) as [HLrw HRrw].
rewrite HLrw in H1.
rewrite HRrw in H2.
clear HLrw HRrw.
intuition.
destruct (classic (exists i, i < n /\ I i /\ J i)).
  (* Case: exists i, i < n /\ I i /\ J i *)
  destruct (K_cond RSk I J s) as [k [HKL [HKR [HKtau HKU]]]] ; try assumption.
  specialize H3 with k s.
  intuition.
    (* Case: k = tau /\ B s s', Contradiction since (k <> tau) (by intuition )*)
    (* Case: exists th t', TauPath (R rk) s' th /\ Trans (R rk) k th t' ... *)
    destruct H3 as [ph [p' [HP [HT HR]]]].
    assert (forall i, i < n -> I i -> J i -> 
              nth i ph dummy = nth i s dummy 
              /\ nth i p' dummy = nth i s dummy) as HRW 
              by (apply HKU ; intuition).
    exists ph.
    apply (TauPath_RSk_then_TauPath_RS RSk I) in HP.
    intuition.
    specialize HRW with i.
    intuition.
  (* Case: ~exists i, i < n /\ I i *)
  exists p.
  split.
  apply rt_refl.
  split ; try exact H.
  intros.
  exfalso.
  apply H4.
  exists i.
  intuition.
Qed.

Lemma TauPath_glue_state_R_prop : 
  forall {n} {RS} {k_fun} {k_fun2} (RSk : @RuleSystemK n RS k_fun k_fun2) (I J : Ensemble nat),
    let SysL (R : RuleSystem) := SystemLTS (LeftNetwork R I) in 
    let SysR (R : RuleSystem) := SystemLTS (RightNetwork R I) in
      forall (B : BB (SysL RSk) (SysR RSk)) s p,
           B s p -> (forall i, i < n -> I i -> J i -> Init (L (nth i (Rules RS) d_tr)) (nthS i p))
            -> exists sh, TauPath (SysL RS) s sh /\ B sh p 
                         /\ (forall i, i < n -> I i -> J i -> nth i p dummy = nth i sh dummy).
Proof.
intros.
destruct (B_states _ _ _ _ _ H).
pose proof (B_branching_sym (SysL RSk) (SysR RSk) B s p H).
unfold SysL, SysR in * ; clear SysL SysR.
destruct (S_cond RSk I) as [HLrw HRrw].
rewrite HLrw in H1.
rewrite HRrw in H2.
clear HLrw HRrw.
intuition.
destruct (classic (exists i, i < n /\ I i /\ J i)).
  (* Case: exists i, i < n /\ I i /\ J i *)
  destruct (K_cond RSk I J p) as [k [HKL [HKR [HKtau HKU]]]] ; try assumption.
  specialize H3 with k p.
  intuition.
    (* Case: k = tau /\ B s s', Contradiction since (k <> tau) (by intuition )*)
    (* Case: exists th t', TauPath (R rk) s' th /\ Trans (R rk) k th t' ... *)
    destruct H3 as [sh [s' [HP [HT HL]]]].
    assert (forall i, i < n -> I i -> J i -> 
              nth i sh dummy = nth i p dummy 
              /\ nth i s' dummy = nth i p dummy) as HRW 
              by (apply HKU ; intuition).
    exists sh.
    apply (TauPath_RSk_then_TauPath_RS RSk I) in HP.
    intuition.
    specialize HRW with i.
    intuition.
  (* Case: ~exists i, i < n /\ I i *)
  exists s.
  split.
  apply rt_refl.
  split ; try exact H.
  intros.
  exfalso.
  apply H4.
  exists i.
  intuition.
Qed.


(* ==================================================================
     Construction of bismulation relations BI
   ================================================================== *)

Lemma BI_rel_states :
  forall {n} {RS} {k_fun} {k_fun2} (RSk : @RuleSystemK n RS k_fun k_fun2) I,
    let SysL (R : RuleSystem) := SystemLTS (LeftNetwork R I) in
    let SysR (R : RuleSystem) := SystemLTS (RightNetwork R I) in
      forall (BI : BB (SysL RSk) (SysR RSk)) s p,
        BI s p -> States (SysL RS) s /\
                  States (SysR RS) p.
Proof.
intros.
pose proof (B_states (SysL RSk) (SysR RSk) BI s p H).
pose proof (S_cond RSk I).
simpl in *.
destruct (H1) as [SLRW SRRW].
rewrite <- SLRW.
rewrite <- SRRW.
auto.
Qed.

Lemma BI_rel_init : 
  forall {n} {RS} {k_fun} {k_fun2} (RSk : @RuleSystemK n RS k_fun k_fun2) I,
    let SysL (R : RuleSystem) := SystemLTS (LeftNetwork R I) in
    let SysR (R : RuleSystem) := SystemLTS (RightNetwork R I) in
      forall (BI : BB (SysL RSk) (SysR RSk)) s, 
           (Init (SysL RS) s -> exists p, Init (SysR RS) p /\ BI s p) 
        /\ (Init (SysR RS) s -> exists p, Init (SysL RS) p /\ BI p s).
Proof.
intros.
pose proof (B_init (SysL RSk) (SysR RSk) BI s).
destruct (I_cond RSk I) as [ILRWs IRRWs].
simpl in *.
rewrite <- ILRWs.
rewrite <- IRRWs.
intuition ;
destruct H6 as [p H6] ;
exists p ;
destruct (I_cond RS k_fun k_fun2 RSk I p) as [ILRWp IRRWp] ;
try rewrite <- IRRWp ;
try rewrite <- ILRWp ;
intuition.
Qed.

Lemma BI_rel_branching : 
  forall {n} {RS} {k_fun} {k_fun2} (RSk : @RuleSystemK n RS k_fun k_fun2) I,
    let SysL (R : RuleSystem) := SystemLTS (LeftNetwork R I) in
    let SysR (R : RuleSystem) := SystemLTS (RightNetwork R I) in
      forall (BI : BB (SysL RSk) (SysR RSk)) s p, BI s p -> 
        forall a s', Trans (SysL RS) a s s' ->
             (a = tau /\ BI s' p)
          \/ (exists ph p', TauPath (SysR RS) p ph /\ Trans (SysR RS) a ph p' 
                         /\ BI s ph /\ BI s' p').
Proof.
intros.
assert (Trans (SysL RSk) a s s').
simpl in *.
unfold NetTrans in *.
destruct H0 as [v HT].
exists v.
intuition.
apply Trans_RS_then_Trans_RSk.
assumption (* End assert *).
pose proof (B_branching (SysL RSk) (SysR RSk) BI s p H a s' H1).
destruct H2.
  (* a = tau *)
  left ; auto.
  (* exists ph p', ... *)
  destruct H2 as [ph [p' HT]] ; auto.
  right ; exists ph, p'.
  intuition.
    (* TauPath (SysR RS) p ph *)
    apply (TauPath_RSk_then_TauPath_RS RSk).
    assumption.
    (* Trans (SysR RS) a ph p' *)
    destruct H4 as [v HT].
    exists v ; intuition.
    apply (Trans_RSk_then_Trans_RS RSk).
    intuition.
    destruct H5 as [Hv _].
    simpl in *.
    destruct Hv as [Hv HK].
    split ; try apply HK.
    destruct Hv ;
      [ apply (SyncL_cond _ RSk x) in H5 ; destruct H5 ; left
      | apply (SyncR_cond _ RSk x) in H5 ; destruct H5 ; right] ;
    try assumption ;
    destruct H5 as [HaRW _] ;
    destruct (trans_cond (SysL RS) a s s') as [_ [_ Ha]] ; auto ;
    destruct (Kact_cond RS k_fun k_fun2 I (sync x)) as [Hna _] ;
    rewrite <- HaRW in Hna ;
    rewrite H4 in Hna ;
    intuition.
Qed.

Lemma BI_rel_branching_sym : 
  forall {n} {RS} {k_fun} {k_fun2} (RSk : @RuleSystemK n RS k_fun k_fun2) I,
    let SysL (R : RuleSystem) := SystemLTS (LeftNetwork R I) in
    let SysR (R : RuleSystem) := SystemLTS (RightNetwork R I) in
      forall (BI : BB (SysL RSk) (SysR RSk)) s p, BI s p -> 
        forall a p', Trans (SysR RS) a p p' ->
             (a = tau /\ BI s p')
          \/ (exists sh s', TauPath (SysL RS) s sh /\ Trans (SysL RS) a sh s' 
                         /\ BI sh p /\ BI s' p').
Proof.
intros.
assert (Trans (SysR RSk) a p p').
simpl in *.
unfold NetTrans in *.
destruct H0 as [v HT].
exists v.
intuition.
apply Trans_RS_then_Trans_RSk.
assumption (* End assert *).
pose proof (B_branching_sym (SysL RSk) (SysR RSk) BI s p H a p' H1).
destruct H2.
  (* a = tau *)
  left ; auto.
  (* exists sh s', ... *)
  destruct H2 as [sh [s' H3]] ; auto.
  right ; exists sh, s'.
  intuition.
    (* TauPath (SysR RS) p ph *)
    apply (TauPath_RSk_then_TauPath_RS RSk I).
    assumption.
    (* Trans (SysR RS) a ph p' *)
    destruct H3 as [v HT].
    exists v ; intuition.
    apply (Trans_RSk_then_Trans_RS RSk I).
    intuition.
    destruct H5 as [Hv _].
    simpl in *.
    destruct Hv as [Hv HF].
    split ; try exact HF.
    apply (SyncL_cond _ RSk v) in Hv.
    destruct Hv.
      (* v not k-sync *)
      exact H5.
      (* v in k-sync *)
      destruct H5 as [HaRW _].
      destruct (trans_cond (SysR RS) a p p') as [_ [_ Ha]] ; auto.
      destruct (Kact_cond RS k_fun k_fun2 I (sync v)) as [_ Hna].
      rewrite <- HaRW in Hna.
      rewrite H3 in Hna.
      intuition.
Qed.

Definition BI_rel
  {n}{RS : RuleSystem}{k_fun : KFun}{k_fun2 : KFun2 RS k_fun}
  (RSk : @RuleSystemK n RS k_fun k_fun2) I 
  (BI : BB (SystemLTS (LeftNetwork (RSk) I)) (SystemLTS (RightNetwork (RSk) I))) :=
    Build_BB n (SystemLTS (LeftNetwork RS I)) 
               (SystemLTS (RightNetwork RS I)) 
               BI
               (BI_rel_states RSk I BI)
               (BI_rel_init RSk I BI)
               (BI_rel_branching RSk I BI)
               (BI_rel_branching_sym RSk I BI).

(* Glue-states are related to themselves *)
Lemma BI_rel_glue_states_related :
  forall {n} {RS} {k_fun} {k_fun2} (RSk : @RuleSystemK n RS k_fun k_fun2) I,
    let SysL (R : RuleSystem) := SystemLTS (LeftNetwork R I) in
    let SysR (R : RuleSystem) := SystemLTS (RightNetwork R I) in
      forall (BI : BB (SysL RSk) (SysR RSk)) s,
        let BI' := BI_rel RSk I BI in
          Init (SysL RS) s -> BI' s s.
Proof.
intros.
destruct (I_cond RSk I) as [HIL HIR].
destruct (B_init (SysL RSk) (SysR RSk) BI s) as [HB _].
unfold SysL in H.
pose proof H.
rewrite <- HIL in H0.
intuition.
simpl.
destruct H1 as [t [HIt HBst]].
destruct (classic (exists i, i < n /\ I i)).
  (* Case: exists i, i < n /\ I i *)
  destruct (KI_cond RSk I s) as [k [HTk [_ [Hk_nottau Hk_unique]]]] ; auto.
  pose proof (B_branching (SysL RSk) (SysR RSk) BI s t HBst k s HTk).
  intuition.
  destruct H3 as [th [t' [_ [HTt Htrest]]]].
  specialize Hk_unique with th t'.
  unfold SysR in HTt.
  intuition.
  rewrite H5 in H2.
  exact H2.
  (* Case: ~exists i, i < n /\ I i *)
  assert (s = t).
    apply (Filter_States_eq (RightNetwork' RSk) I).
    destruct (Glue_cond RSk I) as [HG _].
    rewrite HG in H0 ; clear HG.
    unfold SysR in HIt.
    apply (init_cond (SysR RSk)) in HIt.
    apply (init_cond (SysR RSk)) in H0.
    intuition.
    exfalso.
    apply H1.
    exists i.
    split ; assumption (* End assert *).
  rewrite <- H2 in HBst.
  exact HBst.
Qed.

End TransNetworkExtension.