Module Transformation.

(* ==================================================================
     LTS Transformation
   ================================================================== *)

Load "Utils.ClassicalExt".
Import ClassicalExt.
Load "LTS.LTSN".
Export LTSN.

(* LTS transformation *)
Record TransRule := {
  L : @LTS 1; (* Left pattern *)
  R : @LTS 1; (* Right pattern *)

  (* Init L = Init R = (States L) ∩ (States R) *)
  glue_cond : Init L = Init R /\ Init L = (IntersectS (States L) (States R))
}.

Lemma Init_L_to_Init_R : forall r, Init (L r) = Init (R r).
Proof.
intro r.
pose proof (glue_cond r).
intuition.
Qed.

Lemma Init_L_to_Intersect_LR : forall r, Init (L r) = (IntersectS (States (L r)) (States (R r))).
Proof.
intro r.
pose proof (glue_cond r).
intuition.
Qed.

(* A  function that produces a unique action per state *)
Record KFun := {
  kfun :> State 1 -> Act;

  kfun_injective : forall s s', kfun s = kfun s' -> s = s';
  kfun_not_tau : forall s, kfun s <> tau
}.

(* The kappa-extended transformation rule *)
Record TransRuleK (r : TransRule)(k_fun : KFun) := {
  ker :> TransRule; (* kappa-extension of r *)

  (* kappa-actions are not in (L r) and (R r) *)
  kact_cond : forall s, ~Acts (L r) (k_fun s) /\ ~Acts (R r) (k_fun s);

  s_cond : States (L ker) = States (L r) /\ States (R ker) = States (R r);
  i_cond : Init (L ker) = Init (L r) /\ Init (R ker) = Init (R r);
  a_cond : forall a, (Acts (L ker) a -> Acts (L r) a \/ exists s, Init (L r) s /\ k_fun s = a)
                  /\ (Acts (L r) a -> Acts (L ker) a)
                  /\ (Acts (R ker) a -> Acts (R r) a \/ exists s, Init (R r) s /\ k_fun s = a)
                  /\ (Acts (R r) a -> Acts (R ker) a);

  t_cond : forall a s s', 
       (Trans (L ker) a s s' <-> Trans (L r) a s s' \/ (k_fun s = a /\ s = s' /\ Init (L r) s))
    /\ (Trans (R ker) a s s' <-> Trans (R r) a s s' \/ (k_fun s = a /\ s = s' /\ Init (R r) s))
}.

Definition NoKAct (r : TransRule)(k_fun : KFun)(a : Act) := 
  forall s, Init (L r) s -> k_fun s <> a.

Lemma Not_kact_then_normal_act :
  forall r k_fun (rk : TransRuleK r k_fun) a,
    (NoKAct r k_fun a) ->
          (Acts (L rk) a -> Acts (L r) a)
       /\ (Acts (R rk) a -> Acts (R r) a).
Proof.
unfold NoKAct.
intros.
destruct (a_cond _ _ rk a) as [HAL [_ [HAR _]]].
pose proof (glue_cond r).
split ;
  intuition ;
  destruct H4 as [s Hs] ;
  specialize H with s ;
  [ | rewrite H2 in H] ;
  intuition.
Qed.

Lemma Trans_rk_then_Trans_r :
  forall r k_fun (rk : TransRuleK r k_fun) a s s',
       (Acts (L r) a /\ Trans (L rk) a s s' -> Trans (L r) a s s')
    /\ (Acts (R r) a /\ Trans (R rk) a s s' -> Trans (R r) a s s').
Proof.
intros.
pose proof (t_cond r k_fun rk a s s').
pose proof (kact_cond r k_fun rk s).
intuition ;
rewrite <- H7 in H8 ;
contradiction.
Qed.

Lemma Trans_rk_then_Trans_r2 :
  forall r k_fun (rk : TransRuleK r k_fun) a s s',
       (NoKAct r k_fun a /\ Trans (L rk) a s s' -> Trans (L r) a s s')
    /\ (NoKAct r k_fun a /\ Trans (R rk) a s s' -> Trans (R r) a s s').
Proof.
intros.
destruct (Trans_rk_then_Trans_r _ _ rk a s s') as [HTL HTR].
pose proof (Not_kact_then_normal_act _ _ rk a).
intuition ;
pose proof (trans_cond _ _ _ _ H4) ;
intuition.
Qed.

Lemma TauPath_rk_then_TauPath_r :
  forall r k_fun (rk : TransRuleK r k_fun) s s', 
       (TauPath (R rk) s s' -> TauPath (R r) s s')
    /\ (TauPath (L rk) s s' -> TauPath (L r) s s').
Proof.
intros.
pose proof (t_cond r k_fun rk).
pose proof (kfun_not_tau k_fun).
intuition ; repeat (
  induction H1 ; 
    [ (* Case: Trans G x y -> Trans G x y *)
      apply rt_step ;
      apply (Trans_rk_then_Trans_r2 r k_fun rk) ;
      split ; try assumption ;
      unfold NoKAct ; intros ;
      autounfold ; apply H0
    | (* Case: TauPath G s s *)
      apply rt_refl 
    | (* Case: TauPath G s sh /\ TauPath G sh s' -> TauPath G s s' *)
      apply (rt_trans _ _ x y) ; intuition 
    ]).
Qed.

Lemma taupath_glue_state_prop : 
  forall {r : TransRule} {k_fun : KFun} (rk : @TransRuleK r k_fun) (B : BB (L rk) (R rk)) s s',
    Init (L r) s -> States (R r) s' -> B s s' -> TauPath (R r) s' s.
Proof.
intros.
pose proof (B_branching (L rk) (R rk) B s s').
intuition.
elim (H3 (k_fun s) s).
  (* Case: k = tau /\ B s s' *)
  pose proof (kfun_not_tau k_fun s).
  intuition.
  (* Case: exists th t', TauPath (R rk) s' th /\ Trans (R rk) k th t' ... *)
  intros.
  destruct H2 as [th [t' [HTP [HT _]]]].
  assert (th = s /\ t' = s).
  split ;
  apply (kfun_injective k_fun) ;
  destruct (t_cond _ _ rk (k_fun s) th t') as [_ HTk] ;
  apply HTk in HT ; clear HTk ;
  pose proof (kact_cond _ _ rk s) ;
  destruct HT ;
  first [pose proof (trans_cond _ _ _ _ H4) | intuition ; rewrite <- H2 | auto ] ;
  intuition.
  destruct H2.
  rewrite <- H2.
  apply (TauPath_rk_then_TauPath_r _ _ rk).
  intuition.
(* Exists Trans (L r) k s s' *)
destruct (t_cond _ _ rk (k_fun s) s s) as [HTk _].
intuition.
Qed.

(* Transformation rule match *)
Record Match (H : LTS)(G : LTS) := {
  m :> State 1 -> State 1 -> Prop;

  (* A match maps all states in H *)
  m_matches : forall p, States H p <-> exists s, m p s;
  
  (* A match maps states in H to states in G *)
  m_states : forall p s, m p s -> States H p /\ States G s;

  (* m is injective *)
  m_injective : forall s1 s2 s', (m s' s1 /\ m s' s2) \/ (m s1 s' /\ m s2 s') -> s1 = s2;

  (* A non-glue state may not match onto an initial state *)
  m_init : forall i, Init G i -> forall s, m s i -> Init H s;
  
  (* The transition labels must match *)
  m_trans : forall a p1 p2, Trans H a p1 p2 -> 
              exists s1 s2, m p1 s1 /\ m p2 s2 /\ Trans G a s1 s2;

  (* Dangling conditions *)
  D_cond : forall s1 s2 p1 a,
       States H p1 /\ ~Init H p1 (* p1 in (States H \ Init H) *)
    /\ States G s2 (* s2 in G *)
    -> (
         (m p1 s1 /\ Trans G a s1 s2 -> exists p2, States H p2 /\ Trans H a p1 p2 /\ m p2 s2)
      /\ (m p1 s1 /\ Trans G a s2 s1 -> exists p2, States H p2 /\ Trans H a p2 p1 /\ m p2 s2)
    )
}.

Lemma D_cond_reverse : forall H G (mr : Match H G) s1 s2 p a,
  Trans G a s1 s2 -> (  (mr p s1 /\ (forall p2, ~mr p2 s2 \/ ~Trans H a p p2))
                     \/ (mr p s2 /\ (forall p1, ~mr p1 s1 \/ ~Trans H a p1 p)))
                     -> Init H p.
Proof.
intros.
destruct (trans_cond _ _ _ _ H0) as [HSs1 [HSs2 _]].
apply NNPP.
unfold not ; intros.
destruct H1 as [[Hmp HNex] | [Hmp HNex]] ;
  apply forall_not_or_2_not_exists_and in HNex ;
apply HNex ;
destruct (m_states _ _ _ _ _ Hmp) as [HSp _] ;
  [ pose proof (D_cond H G mr s1 s2 p a)
  | pose proof (D_cond H G mr s2 s1 p a) ] ;
intuition ;
destruct H3 as [px Hpx] ;
exists px ;
intuition.
Qed.

Definition image_mr {H : LTS}{G : LTS}(mr : Match H G)(S : Ensemble (State 1))(s : State 1) :=
  (exists p, mr p s /\ S p).

Definition matched {H : LTS}{G : LTS}(mr : Match H G)(s : State 1) := exists p, mr p s.

Record Transform {r : TransRule}{G : LTS}(mL : Match (L r) G) := {
  TG :> LTS; (* Transformed G *)

  mR : Match (R r) TG;
  mR_init_cond : forall q, Init (L r) q -> mL q = mR q;

  (* The matches do not match on states on the opposite side of the transformation *)
  mR_G_cond : forall q p, mR q p -> States G p  -> Init (L r) q;
  mL_G_cond : forall q p, mL q p -> States TG p -> Init (R r) q;

  IT_cond : Init TG = Init G;
  
  ST_cond : States TG = UnionS (SetminusS (States G) (matched mL)) (matched mR);

  TT_cond :
      forall a s s', 
        Trans TG a s s' <-> (
             (* Remove transitions that are matched by mL *)
             (Trans G a s s' /\ ~(exists q q', mL q s /\ mL q' s' /\ Trans (L r) a q q'))
          \/ (* Add transition thar are matched by mR *)
             exists q q', mR q s /\ mR q' s' /\ Trans (R r) a q q'
         );

  AT_cond : forall a, Acts TG a <-> (a = tau \/ exists s s', Trans TG a s s')
}.

Lemma Trans_R_in_TG : 
  forall r G (mL : Match (L r) G) (TrG : Transform mL),
    let mR := (mR mL TrG) in
      forall a p1 p2, Trans (R r) a p1 p2 ->
        exists s1 s2, mR p1 s1 /\ mR p2 s2 /\ Trans TrG a s1 s2.
Proof.
intros.
destruct (m_trans (R r) TrG mR0 a p1 p2 H) as [t1 [t2 HT]].
exists t1, t2.
auto.
Qed.

(* ==================================================================
     Dummy TransRule
   ================================================================== *)

(* Dummy transformation rule *)
Lemma Dummy_glue_cond : (Init Singleton_LTS) = (Init Singleton_LTS) 
  /\ (Init Singleton_LTS) = (IntersectS (Init Singleton_LTS) (Init Singleton_LTS)).
Proof.
unfold IntersectS.
intuition.
apply Extensionality_Ensembles.
unfold Same_set.
unfold Included.
intuition.
destruct H.
assumption.
Qed.

Definition DummyTransRule := Build_TransRule Singleton_LTS Singleton_LTS Dummy_glue_cond.

Notation d_lts := Singleton_LTS.
Notation d_tr  := DummyTransRule.

Lemma d_lts_Ltr : d_lts = L d_tr.
Proof.
simpl ; reflexivity.
Qed.

End Transformation.