(* LTS transformation network and related notions *)

Module TransformationNetwork.
Load "LTS.Network".
Export Network.
Load "Utils.OptionProps".
Import OptionProps.
Load "Utils.DomainSplitting".
Import DomainSplitting.
Load "Utils.ClassicalExt".
Import ClassicalExt.

(* ==================================================================
     System of transformation rules
   ================================================================== *)

(* System of transformation rules *)
Record RuleSystem {n} := {
  Rules : listn TransRule n; (* Set of transformation rules *)

  SyncL : Ensemble (@SyncLaw n); (* Set of existing synchronisation laws *)
  SyncR : Ensemble (@SyncLaw n); (* Set of (new) synchronisation laws *)

  AdmL_cond : Admissible SyncL (map L Rules);
  AdmR_cond : Admissible SyncR (map R Rules)
}.

Lemma RS_AdmLR : forall {n}(RS : @RuleSystem n), 
  Admissible (Union _ (SyncL RS) (SyncR RS)) (map R (Rules RS)).
Proof.
intros.
pose proof (AdmL_cond RS) as HL.
pose proof (AdmR_cond RS) as HR.
unfold Admissible, NoSyncAndRename, NoCut in *.
split ; intros.
destruct H as [HV [Hin Hv]].
destruct HV ; [apply HL | apply HR] ; intuition.
destruct HR as [_ HR].
destruct (HR i H) as [v [HVv Hvtau]].
exists v.
split ; try assumption.
apply Union_intror.
exact HVv.
Qed.

Ltac F_map_nth F := intros ; assert (F d_tr = d_lts) as HFrw by (simpl ; reflexivity) ; 
                    rewrite <- HFrw ; apply map_nth.

Lemma L_map_nth : 
  forall {n} (RS : @RuleSystem n) i, 
    nth i (map L (Rules RS)) d_lts = L (nth i (Rules RS) d_tr).
Proof.
F_map_nth L.
Qed.

Lemma R_map_nth : 
  forall {n} (RS : @RuleSystem n) i, 
    nth i (map R (Rules RS)) d_lts = R (nth i (Rules RS) d_tr).
Proof.
F_map_nth R.
Qed.

Lemma F_map_length : 
  forall {n} (RS : @RuleSystem n) (F : TransRule -> (@LTS 1)), 
    length (map F (Rules RS)) = n.
Proof.
intros.
rewrite map_length.
rewrite lst_length.
reflexivity.
Qed.

(* The left pattern network *)
Definition LeftNetwork' {n}(RS : RuleSystem) :=
  Build_LTSNetwork n (Bld_listn (map L (Rules RS)) (F_map_length RS L)) 
                     (SyncL RS) (AdmL_cond RS).

(* The right pattern network *)					 
Definition RightNetwork' {n}(RS : RuleSystem) := 
  Build_LTSNetwork n (Bld_listn (map R (Rules RS)) (F_map_length RS R))
                     (Union SyncLaw (SyncL RS) (SyncR RS)) (RS_AdmLR RS).

(* The left and right filtered pattern networks *)
Definition LeftNetwork {n}(RS : @RuleSystem n)(I : Ensemble nat) := FilterNetwork (LeftNetwork' RS) I.
Definition RightNetwork {n}(RS : @RuleSystem n)(I : Ensemble nat) := FilterNetwork (RightNetwork' RS) I.

Lemma LeftNetwork_Procs_i_in_I : 
  forall {n} (RS : @RuleSystem n) (I : Ensemble nat) i, i < n -> I i ->
    (nth i (Procs (LeftNetwork RS I)) d_lts) = (L (nth i (Rules RS) d_tr)).
Proof.
intros.
unfold LeftNetwork.
rewrite (FilterNetwork_Procs_i_in_I _ I i H H0).
rewrite <- (L_map_nth RS i).
simpl.
reflexivity.
Qed.

Lemma RightNetwork_Procs_i_in_I : 
  forall {n} (RS : @RuleSystem n) (I : Ensemble nat) i, i < n -> I i ->
    (nth i (Procs (RightNetwork RS I)) d_lts) = (R (nth i (Rules RS) d_tr)).
Proof.
intros.
unfold RightNetwork.
rewrite (FilterNetwork_Procs_i_in_I _ I i H H0).
rewrite <- (R_map_nth RS i).
simpl.
reflexivity.
Qed.

Lemma LeftNetwork_States_in_I : 
  forall {n} (RS : @RuleSystem n) I s, States (SystemLTS (LeftNetwork RS I)) s -> 
    forall i, i < n -> I i -> States (L (nth i (Rules RS) d_tr)) (nthS i s).
Proof.
intros.
rewrite <- L_map_nth.
apply (FilterNetwork_States_in_I _ I s H i H0 H1).
Qed.

Lemma RightNetwork_States_in_I : 
  forall {n} (RS : @RuleSystem n) I s, States (SystemLTS (RightNetwork RS I)) s -> 
    forall i, i < n -> I i -> States (R (nth i (Rules RS) d_tr)) (nthS i s).
Proof.
intros.
rewrite <- R_map_nth.
apply (FilterNetwork_States_in_I _ I s H i H0 H1).
Qed.

Lemma LeftNetwork_States_simpl :
  forall {n} (RS : @RuleSystem n) I s i, States (SystemLTS (LeftNetwork RS I)) s ->
    (i < n -> 
         (I i -> States (L (nth i (Rules RS) d_tr)) (nthS i s))
      /\ (~I i -> nthS i s = d_state)).
Proof.
intros.
pose proof (FilterNetwork_States (LeftNetwork' RS) I s).
rewrite <- L_map_nth.
apply H1 ; assumption.
Qed.

Lemma LeftNetwork_States_lift :
  forall {n} (RS : @RuleSystem n) (I : Ensemble nat) s, 
    (forall i, i < n -> 
         (I i -> States (L (nth i (Rules RS) d_tr)) (nthS i s))
      /\ (~I i -> nthS i s = d_state)) -> States (SystemLTS (LeftNetwork RS I)) s.
Proof.
intros.
apply (FilterNetwork_States (LeftNetwork' RS) I s).
intuition ; simpl ;
try rewrite L_map_nth ;
apply H ; assumption.
Qed.

Lemma RightNetwork_States_simpl :
  forall {n} (RS : @RuleSystem n) I s i, States (SystemLTS (RightNetwork RS I)) s ->
    (i < n ->   (I i -> States (R (nth i (Rules RS) d_tr)) (nthS i s))
             /\ (~I i -> nthS i s = d_state)).
Proof.
intros.
pose proof (FilterNetwork_States (RightNetwork' RS) I s).
rewrite <- R_map_nth.
apply H1 ; assumption.
Qed.

Lemma RightNetwork_States_lift :
  forall {n} (RS : @RuleSystem n) (I : Ensemble nat) s, 
    (forall i, i < n ->   (I i -> States (R (nth i (Rules RS) d_tr)) (nthS i s))
                       /\ (~I i -> nthS i s = d_state)) -> States (SystemLTS (RightNetwork RS I)) s.
Proof.
intros.
apply (FilterNetwork_States (RightNetwork' RS) I s).
intuition ; simpl ;
try rewrite R_map_nth ;
apply H ; assumption.
Qed.

Lemma LeftNetwork_Init_simpl :
  forall {n} (RS : @RuleSystem n) I s i, 
    Init (SystemLTS (LeftNetwork RS I)) s ->
    (i < n ->   (I i -> Init (L (nth i (Rules RS) d_tr)) (nthS i s))
             /\ (~I i -> nthS i s = d_state)).
Proof.
intros.
pose proof (FilterNetwork_Init (LeftNetwork' RS) I s).
rewrite <- L_map_nth.
apply H1 ; assumption.
Qed.

Lemma LeftNetwork_Init_lift :
  forall {n} (RS : @RuleSystem n) (I : Ensemble nat) s, 
    (forall i, i < n ->   (I i -> Init (L (nth i (Rules RS) d_tr)) (nthS i s))
                       /\ (~I i -> nthS i s = d_state)) -> Init (SystemLTS (LeftNetwork RS I)) s.
Proof.
intros.
apply (FilterNetwork_Init (LeftNetwork' RS) I s).
intuition ; simpl ;
try rewrite L_map_nth ;
apply H ; assumption.
Qed.

Lemma RightNetwork_Init_simpl :
  forall {n} (RS : @RuleSystem n) I s i, Init (SystemLTS (RightNetwork RS I)) s ->
    (i < n ->   (I i -> Init (R (nth i (Rules RS) d_tr)) (nthS i s))
             /\ (~I i -> nthS i s = d_state)).
Proof.
intros.
pose proof (FilterNetwork_Init (RightNetwork' RS) I s).
rewrite <- R_map_nth.
apply H1 ; assumption.
Qed.

Lemma RightNetwork_Init_lift :
  forall {n} (RS : @RuleSystem n) (I : Ensemble nat) s, 
    (forall i, i < n ->   (I i -> Init (R (nth i (Rules RS) d_tr)) (nthS i s))
                       /\ (~I i -> nthS i s = d_state)) -> Init (SystemLTS (RightNetwork RS I)) s.
Proof.
intros.
apply (FilterNetwork_Init (RightNetwork' RS) I s).
intuition ; simpl ;
try rewrite R_map_nth ;
apply H ; assumption.
Qed.

Lemma Glue_cond : 
  forall {n}(RS : @RuleSystem n) I,
    let SysL := SystemLTS (LeftNetwork RS I) in
    let SysR := SystemLTS (RightNetwork RS I) in
      Init SysL = Init SysR /\ Init SysL = (IntersectS (States SysL) (States SysR)).
Proof.
intros.
split ; apply Extensionality_Ensembles ;
unfold Same_set, Included ;
split ; intros ;
unfold In in * ;
intros ;
[ apply RightNetwork_Init_lift
| apply LeftNetwork_Init_lift
| split ; [apply init_cond | apply RightNetwork_States_lift]
| apply LeftNetwork_Init_lift] ;
try intros i Hin ;
[ apply (LeftNetwork_Init_simpl RS I x i) in H 
| apply (RightNetwork_Init_simpl RS I x i) in H 
| | apply (LeftNetwork_Init_simpl RS I x i) in H
| rewrite Init_L_to_Intersect_LR
]; try exact Hin ; intuition ; try apply init_cond ;
try (rewrite Init_L_to_Init_R || rewrite <- Init_L_to_Init_R) ; try exact H ; try exact H2 ;
(* Case: Included (IntersectS (States SysL) (States SysR)) (Init SysL) *)
destruct H ;
apply (LeftNetwork_States_simpl RS I x i) in H ;
apply (RightNetwork_States_simpl RS I x i) in H1 ; 
try exact Hin ;
try split ; intuition.
Qed.

Lemma Net_Init_L_to_Init_R : 
  forall {n}(RS : @RuleSystem n) I,
      Init (SystemLTS (LeftNetwork RS I)) = Init (SystemLTS (RightNetwork RS I)).
Proof.
intros n RS I.
destruct (Glue_cond RS I).
rewrite H.
reflexivity.
Qed.

(* ==================================================================
     LTS network transformation
   ================================================================== *)

(* Specification of network transformation *)
Record NetTransform {n}{M : @LTSNetwork n}{RS : @RuleSystem n} := {
  TM :> LTSNetwork;

  (* Assumption that for each i a match exists *)
  mLs (i : nat) : Match (L (nth i (Rules RS) d_tr)) (nth i (Procs M) d_lts);

  (* A RuleSystem can only be applied if the left-patterns sync laws match those of the network *)
  SyncMatch_cond : SyncL RS = SyncN M;
  SyncN_cond : SyncN TM = Union SyncLaw (SyncN M) (SyncR RS);

  TrG (i : nat) : Transform (mLs i);
  Procs_cond : forall i, i < n -> nth i (Procs TM) d_lts = TrG i
}.

(* ------------------------------------------------------------------
     Properties of the active set of matched composed states
   ------------------------------------------------------------------ *)

Lemma Ac_not_removed_InitL :
  forall {n}{M : LTSNetwork}{RS : RuleSystem}(TrM : @NetTransform n M RS)(v : @SyncLaw n)(s' p p' : State n) i,
    States (SystemLTS TrM) s' -> p' = Cmps p s' (Ac v) -> Ac v i ->
      forall q', (mLs TrM i) q' (nthS i p') -> Init (L (nth i (Rules RS) d_tr)) q'.
Proof.
intros n M RS TrM v s' p p' i HSs' Hcmps Hv q' Hmq'.
pose proof (Cmps_In p s' (Ac v)) as Hp'.
rewrite <- Hcmps in Hp'.
rewrite Init_L_to_Init_R.
apply (mL_G_cond _ (TrG TrM i) _ _ Hmq').
pose proof (Acv_i_st_n v i Hv) as Hin.
destruct (Hp' i Hin) as [Hp'Ac _].
unfold nthS ; rewrite (Hp'Ac Hv).
rewrite <- Procs_cond ; try exact Hin.
apply HSs'; exact Hin.
Qed.

Lemma Ac_not_removed_InitR :
  forall {n}{M : LTSNetwork}{RS : RuleSystem}(TrM : @NetTransform n M RS)(v : @SyncLaw n)(s' p p' : State n) i,
    States (SystemLTS M) s' -> p' = Cmps p s' (Ac v) -> Ac v i ->
      forall q', (mR _ (TrG TrM i)) q' (nthS i p') -> Init (R (nth i (Rules RS) d_tr)) q'.
Proof.
intros n M RS TrM v s' p p' i HSs' Hcmps Hv q' Hmq'.
pose proof (Cmps_In p s' (Ac v)) as Hp'.
rewrite <- Hcmps in Hp'.
rewrite <- Init_L_to_Init_R.
apply (mR_G_cond _ _ _ _ Hmq').
pose proof (Acv_i_st_n v i Hv) as Hin.
destruct (Hp' i Hin) as [Hp'Ac _].
unfold nthS ; rewrite (Hp'Ac Hv) ; apply HSs'; exact Hin.
Qed.

(* -------------------------------------------------------------------------------------------
     The set of indices of elements in s matched on by the corresponding transformation rule
   ------------------------------------------------------------------------------------------- *)

Definition MatchIds {n}{A}{M : LTSNetwork}{RS : RuleSystem} (N : TransRule -> LTS)(TrM : @NetTransform n M RS)
                       (s : State n)(i : nat)(mr : Match (N (nth i (Rules RS) d_tr)) A) :=
  i < n /\ matched mr (nthS i s).

Definition MatchLIds {n}{M : LTSNetwork}{RS : RuleSystem}(TrM : @NetTransform n M RS)
                       (s : State n)(i : nat) :=
  MatchIds L TrM s i (mLs TrM i).

Definition MatchRIds {n}{M : LTSNetwork}{RS : RuleSystem}(TrM : @NetTransform n M RS)
                       (s : State n)(i : nat) :=
  MatchIds R TrM s i (mR _ (TrG TrM i)).

(* ==================================================================
     Mapping of state vectors
   ================================================================== *)

Definition MapL {n}{M : LTSNetwork}{RS : RuleSystem}(TrM : @NetTransform n M RS)
                (I : Ensemble nat)(p s : State n) :=
       States (SystemLTS M) s /\ States (SystemLTS (LeftNetwork RS I)) p
    /\ forall i, i < n -> I i -> mLs TrM i (nthS i p) (nthS i s).

Definition MapR {n}{M : LTSNetwork}{RS : RuleSystem}(TrM : @NetTransform n M RS)
                (I : Ensemble nat)(p s : State n) :=
       States (SystemLTS TrM) s /\ States (SystemLTS (RightNetwork RS I)) p
    /\ forall i, i < n -> I i -> mR _ (TrG TrM i) (nthS i p) (nthS i s).

Lemma MapL_maps : 
  forall {n} {M} {RS} (TrM : @NetTransform n M RS) I p,
    States (SystemLTS (LeftNetwork RS I)) p <-> exists s, MapL TrM I p s.
Proof.
intros.
unfold MapL.
split ; intro.
(* States NetL p -> exists s, MapL TrM I p s *)
apply exists_forall_domain3.
set (P := fun a s 
            =>  States (nth a (Procs M) d_lts) (Bld_list1 s)
             /\ States (nth a (Procs (LeftNetwork RS I)) d_lts) (nthS a p)
             /\ (I a -> (mLs TrM a) (nthS a p) (Bld_list1 s))).
apply (listn_forall_exists _ n dummy P).
unfold P ; clear P.
intros.
destruct (EnsembleNat_dec I i) as [HI | HI].
  (* I i *)
  pose proof (LeftNetwork_States_in_I RS I p H i H0 HI) as HSpi.
  apply (m_matches _ _ (mLs TrM i) (nthS i p) ) in HSpi.
  destruct HSpi as [s Hs].
  exists (hd dummy s).
  rewrite list1_prop.
  pose proof (m_states _ _ (mLs TrM i) _ _ Hs).
  intuition.
  (* ~I i *)
  destruct (init_not_empty_cond (nth i (Procs M) d_lts)) as [s Hs].
  exists (hd dummy s).
  rewrite list1_prop.
  intuition.
  apply init_cond.
  exact Hs.
(* exists s, MapL TrM I p s -> States NetL p *)
intro.
destruct H as [_ [_ [HLS _]]].
apply HLS.
Qed.

Lemma MapR_maps : 
  forall {n} {M} {RS} (TrM : @NetTransform n M RS) I p,
    States (SystemLTS (RightNetwork RS I)) p <-> exists s, MapR TrM I p s.
Proof.
intros.
unfold MapR.
split ; intro.
(* States NetR p -> exists s, MapR TrM I p s *)
apply exists_forall_domain3.
set (P := fun a s 
            =>  States (nth a (Procs TrM) d_lts) (Bld_list1 s)
             /\ States (nth a (Procs (RightNetwork RS I)) d_lts) (nthS a p)
             /\ (I a -> (mR (mLs TrM a) (TrG TrM a)) (nthS a p) (Bld_list1 s))).
apply (listn_forall_exists _ n dummy P).
unfold P ; clear P.
intros.
destruct (EnsembleNat_dec I i) as [HI | HI].
  (* I i *)
  pose proof (RightNetwork_States_in_I RS I p H i H0 HI) as HSpi.
  apply (m_matches _ _ (mR (mLs TrM i) (TrG TrM i)) (nthS i p)) in HSpi.
  destruct HSpi as [s Hs].
  exists (hd dummy s).
  rewrite list1_prop.
  pose proof (m_states _ _ (mR _ (TrG TrM i)) _ _ Hs).
  rewrite (Procs_cond TrM); try exact H0.
  intuition.
  (* ~I i *)
  destruct (init_not_empty_cond (nth i (Procs TrM) d_lts)) as [s Hs].
  exists (hd dummy s).
  rewrite list1_prop.
  intuition.
  apply init_cond.
  exact Hs.
(* exists s, MapL TrM I p s -> States NetL p *)
intro.
destruct H as [_ [_ [HLS _]]].
apply HLS.
Qed.

Lemma MapL_states : 
  forall {n} {M} {RS} (TrM : @NetTransform n M RS) I p s, 
    MapL TrM I p s ->   States (SystemLTS (LeftNetwork RS I)) p 
                     /\ States (SystemLTS M) s.
Proof.
unfold MapL.
intuition.
Qed.

Lemma MapR_states : 
  forall {n} {M} {RS} (TrM : @NetTransform n M RS) I p s, 
    MapR TrM I p s ->   States (SystemLTS (RightNetwork RS I)) p 
                     /\ States (SystemLTS TrM) s.
Proof.
unfold MapR.
intuition.
Qed.

Lemma MapL_injective :
  forall {n} {M} {RS} (TrM : @NetTransform n M RS) I p1 p2 s,
    MapL TrM I p1 s -> MapL TrM I p2 s -> p1 = p2.
Proof.
unfold MapL.
intros.
apply (Filter_States_eq (LeftNetwork' RS) I).
intuition.
specialize H4 with i.
specialize H5 with i.
intuition.
pose proof (m_injective _ _ (mLs TrM i)).
apply H4 with (s' := (nthS i s)).
intuition.
Qed.

Lemma MapR_injective :
  forall {n} {M} {RS} (TrM : @NetTransform n M RS) I p1 p2 s,
    MapR TrM I p1 s -> MapR TrM I p2 s -> p1 = p2.
Proof.
unfold MapR.
intros.
apply (Filter_States_eq (RightNetwork' RS) I).
intuition.
specialize H4 with i.
specialize H5 with i.
intuition.
pose proof (m_injective _ _ (mR (mLs TrM i) (TrG TrM i))).
apply H4 with (s' := (nthS i s)).
intuition.
Qed.

(* A non-glue state may not map onto an initial state *)
Lemma MapL_init : 
  forall {n} {M} {RS} (TrM : @NetTransform n M RS) I s, 
    Init (SystemLTS M) s -> 
      forall p, MapL TrM I p s -> 
        Init (SystemLTS (LeftNetwork RS I)) p.
Proof.
unfold MapL.
intuition.
apply (FilterNetwork_Init (LeftNetwork' RS) I p).
intuition.
simpl ; rewrite L_map_nth.
apply (m_init _ _ (mLs TrM i) (nthS i s) (H i H2)).
apply H3; assumption.
apply (FilterNetwork_States_not_in_I (LeftNetwork' RS) I) ; assumption.
Qed.

Lemma MapR_init : 
  forall {n} {M} {RS} (TrM : @NetTransform n M RS) I s, 
    Init (SystemLTS TrM) s -> 
      forall p, MapR TrM I p s -> 
        Init (SystemLTS (RightNetwork RS I)) p.
Proof.
unfold MapR.
intuition.
apply (FilterNetwork_Init (RightNetwork' RS) I p).
intuition.
simpl ; rewrite R_map_nth.
simpl in H.
unfold NetInit in H.
specialize H with i.
rewrite Procs_cond in H ; try assumption.
apply (m_init _ _ (mR _ (TrG TrM i)) (nthS i s) (H H2)).
apply H3; assumption.
apply (FilterNetwork_States_not_in_I (RightNetwork' RS) I) ; assumption.
Qed.

Lemma Map_trans_rewrite_helper : 
  forall {A} n (P Q R S T : nat -> A -> Prop) (U : Prop),
    (U /\ exists s2, forall i, i < n -> 
       P i s2 /\ Q i s2 /\ R i s2 /\ S i s2 /\ T i s2) ->
    (exists s2,
    ((forall i : nat, i < n -> P i s2) /\
      (forall i : nat, i < n -> Q i s2) /\
      (forall i : nat, i < n -> R i s2)) /\
    (U /\
      (forall i : nat, i < n -> S i s2)) /\
      (forall i : nat, i < n -> T i s2) ).
Proof.
intuition.
destruct H1 as [s2 Hi].
exists s2.
intuition ; specialize Hi with i ; intuition.
Qed.

Lemma MapL_trans : 
  forall {n} {M} {RS} (TrM : @NetTransform n M RS) I v p1 p2 s1,
    let NetL := LeftNetwork RS I in
      NetTransSL NetL v p1 p2 -> MapL TrM I p1 s1 -> 
        exists s2, MapL TrM I p2 s2 /\ NetTransSL M v s1 s2
          /\ forall i, i < n -> ~I i -> nthS i s1 = nthS i s2.
Proof.
unfold MapL.
simpl.
unfold NetTransSL, NetStates.
intros.
apply Map_trans_rewrite_helper.
intuition.
simpl in H1.
destruct H1 as [H1 _].
rewrite <- (SyncMatch_cond (RS:=RS)) ; assumption.
set (P := fun i s2
   => States (nth i (Procs M) d_lts) (Bld_list1 s2) /\
      States (nth i (Procs (LeftNetwork RS I)) d_lts) (nthS i p2) /\
      (I i -> (mLs TrM i) (nthS i p2) (Bld_list1 s2)) /\
      match nth i (sync v) None with
      | Some x => Trans (nth i (Procs M) d_lts) x (nthS i s1) (Bld_list1 s2)
      | None => nth i s1 dummy = s2 /\
                States (nth i (Procs M) d_lts) (nthS i s1) /\
                States (nth i (Procs M) d_lts) (Bld_list1 s2) end /\
      ((I i -> False) -> nthS i s1 = (Bld_list1 s2)) ).
apply (listn_forall_exists _ n dummy P).
unfold P ; clear P (* End rewrite *).
intros.
specialize H with i.
specialize H2 with i.
specialize H4 with i.
intuition.
simpl in H5.
simpl.
rewrite Filter_nth in H5.
rewrite Filter_nth.
destruct (Ensemble2boollist_In n I i H3) as [[_ HIn] [_ HNIn]].
destruct (EnsembleNat_dec I i) as [HI | HI] ; 
  [rewrite (HIn HI) in * | rewrite (HNIn HI) in *] ; auto ; simpl in *.
  (* I i *)
  rewrite L_map_nth in H5 ; rewrite L_map_nth.
  destruct (nth i (sync v) None).
    (* Ac v i *)
    apply (m_trans _ _ (mLs TrM i)) in H5.
    destruct H5 as [s1i [s2i [Hm1 [Hm2 HT]]]].
    exists (hd dummy s2i).
    rewrite list1_prop.
    pose proof (m_states _ _ (mLs TrM i) (nthS i p2) s2i).
    pose proof (m_injective _ _ (mLs TrM i) (nthS i s1) s1i (nthS i p1)).
    intuition.
    rewrite H5.
    exact HT.
    (* ~Ac v i *)
    destruct H5 as [HRW [_ HSp2]].
    unfold nthS in H.
    rewrite HRW in H.
    exists (nth i s1 dummy).
    intuition.
  (* ~I i *)
  exists (nth i s1 dummy).
  destruct (nth i (sync v) None) ; intuition.
Qed.

Lemma MapR_trans : 
  forall {n} {M} {RS} (TrM : @NetTransform n M RS) I v p1 p2 s1,
    let NetR := RightNetwork RS I in
      NetTransSL NetR v p1 p2 -> MapR TrM I p1 s1 -> 
        exists s2, MapR TrM I p2 s2 /\ NetTransSL TrM v s1 s2 
                /\ forall i, i < n -> ~I i -> nthS i s1 = nthS i s2.
Proof.
unfold MapR.
simpl.
unfold NetTransSL, NetStates.
intros.
apply Map_trans_rewrite_helper.
intuition.
simpl in H1.
rewrite (SyncN_cond TrM).
destruct H1 as [[H1 | H1] _].
left ; rewrite <- (SyncMatch_cond (RS:=RS)) ; assumption.
right ; assumption.
set (P := fun i s2
   => States (nth i (Procs TrM) d_lts) (Bld_list1 s2) /\
      States (nth i (Procs (RightNetwork RS I)) d_lts) (nthS i p2) /\
      (I i -> (mR (mLs TrM i) (TrG TrM i)) (nthS i p2) (Bld_list1 s2)) /\
      match nth i (sync v) None with
      | Some x => Trans (nth i (Procs TrM) d_lts) x (nthS i s1) (Bld_list1 s2)
      | None => nth i s1 dummy = s2 /\
                States (nth i (Procs TrM) d_lts) (nthS i s1) /\
                States (nth i (Procs TrM) d_lts) (Bld_list1 s2) end /\
      ((I i -> False) -> nthS i s1 = (Bld_list1 s2)) ).
pose proof (listn_forall_exists _ n dummy P).
apply H3.
unfold P ; clear H3 P (* End rewrite *).
intros.
specialize H with i.
specialize H2 with i.
specialize H4 with i.
intuition.
simpl in H5.
simpl.
rewrite Filter_nth in H5.
rewrite Filter_nth.
destruct (Ensemble2boollist_In n I i H3) as [[_ HIn] [_ HNIn]].
destruct (EnsembleNat_dec I i) as [HI | HI] ; 
  [rewrite (HIn HI) in * | rewrite (HNIn HI) in *] ; auto ; simpl in *.
  (* I i *)
  rewrite R_map_nth in H5 ; rewrite R_map_nth.
  destruct (nth i (sync v) None).
    (* Ac v i *)
    rewrite (Procs_cond TrM i H3).
    apply (m_trans _ _ (mR (mLs TrM i) (TrG TrM i))) in H5.
    destruct H5 as [s1i [s2i [Hm1 [Hm2 HT]]]].
    exists (hd dummy s2i).
    rewrite list1_prop.
    pose proof (m_states _ _ (mR (mLs TrM i) (TrG TrM i)) (nthS i p2) s2i).
    pose proof (m_injective _ _ (mR (mLs TrM i) (TrG TrM i)) (nthS i s1) s1i (nthS i p1)).
    intuition.
    rewrite H5.
    exact HT.
    (* ~Ac v i *)
    destruct H5 as [HRW [_ HSp2]].
    unfold nthS in H.
    rewrite HRW in H.
    exists (nth i s1 dummy).
    intuition.
  (* ~I i *)
  exists (nth i s1 dummy).
  destruct (nth i (sync v) None) ; intuition.
Qed.

Lemma MapL_trans2 :
  forall {n}{M}{RS}(TrM : @NetTransform n M RS) (I : Ensemble nat) a p1 p2 s1,
    Trans (SystemLTS (LeftNetwork RS I)) a p1 p2 -> MapL TrM I p1 s1 ->
      exists s2, MapL TrM I p2 s2 /\ Trans (SystemLTS M) a s1 s2
             /\ forall i, i < n -> ~I i -> nthS i s1 = nthS i s2.
Proof.
simpl.
unfold NetTrans.
intros.
destruct H as [v [Hva HT]].
destruct (MapL_trans TrM I v p1 p2 s1 HT H0) as [s2 HX].
exists s2.
intuition.
exists v.
intuition.
Qed.

Lemma MapR_trans2 :
  forall {n}{M}{RS}(TrM : @NetTransform n M RS) (I : Ensemble nat) a p1 p2 s1,
    Trans (SystemLTS (RightNetwork RS I)) a p1 p2 -> MapR TrM I p1 s1 ->
      exists s2, MapR TrM I p2 s2 /\ Trans (SystemLTS TrM) a s1 s2
             /\ forall i, i < n -> ~I i -> nthS i s1 = nthS i s2.
Proof.
simpl.
unfold NetTrans.
intros.
destruct H as [v [Hva HT]].
destruct (MapR_trans TrM I v p1 p2 s1 HT H0) as [s2 HX].
exists s2.
intuition.
exists v.
intuition.
Qed.

Lemma MapL_TauPath :
  forall {n}{M}{RS}(TrM : @NetTransform n M RS) (I : Ensemble nat) p1 p2,
    TauPath (SystemLTS (LeftNetwork RS I)) p1 p2 -> 
      forall s1, MapL TrM I p1 s1 -> 
        exists s2, MapL TrM I p2 s2 /\ TauPath (SystemLTS M) s1 s2
                /\ forall i, i < n -> ~I i -> nthS i s1 = nthS i s2.
Proof.
intros n M RS TrM I p1 p2 HT.
induction HT ; intros.
  (* rt_step *)
  pose proof (MapL_trans2 TrM I tau x y s1 H H0).
  destruct H1 as [s2 [HMs2 [HT Heq]]].
  exists s2.
  split ; try exact HMs2.
  split ; try exact Heq.
  apply rt_step.
  exact HT.  
  (* rt_refl *)
  exists s1.
  split ; try exact H.
  intuition.
  apply rt_refl.
  (* rt_trans *)
  destruct (IHHT1 s1 H) as [sy [HMy [HTy Heqy]]].
  destruct (IHHT2 sy HMy) as [s2 [HMz [HTz Heqz]]].
  exists s2.
  split ; try exact HMz.
  split.
  apply (rt_trans _ _ s1 sy s2) ; assumption.
  intros.
  rewrite (Heqy i H0 H1).
  exact (Heqz i H0 H1).
Qed.

Lemma MapR_TauPath :
  forall {n}{M}{RS}(TrM : @NetTransform n M RS) (I : Ensemble nat) p1 p2,
    TauPath (SystemLTS (RightNetwork RS I)) p1 p2 -> 
      forall s1, MapR TrM I p1 s1 -> 
        exists s2, MapR TrM I p2 s2 /\ TauPath (SystemLTS TrM) s1 s2
                /\ forall i, i < n -> ~I i -> nthS i s1 = nthS i s2.
Proof.
intros n M RS TrM I p1 p2 HT.
induction HT ; intros.
  (* rt_step *)
  pose proof (MapR_trans2 TrM I tau x y s1 H H0).
  destruct H1 as [s2 [HMs2 [HT Heq]]].
  exists s2.
  split ; try exact HMs2.
  split ; try exact Heq.
  apply rt_step.
  exact HT.  
  (* rt_refl *)
  exists s1.
  split ; try exact H.
  intuition.
  apply rt_refl.
  (* rt_trans *)
  destruct (IHHT1 s1 H) as [sy [HMy [HTy Heqy]]].
  destruct (IHHT2 sy HMy) as [s2 [HMz [HTz Heqz]]].
  exists s2.
  split ; try exact HMz.
  split.
  apply (rt_trans _ _ s1 sy s2) ; assumption.
  intros.
  rewrite (Heqy i H0 H1).
  exact (Heqz i H0 H1).
Qed.

Lemma MapL_D_reverse :
  forall {n} {M} {RS} (TrM : @NetTransform n M RS) I s1 s2 p1 v,
    let Mp := MapL TrM I in
    let Li (i : nat) := L (nth i (Rules RS) d_tr) in
      Mp p1 s1 ->
      (    (NetTransSL M v s1 s2 /\
             (forall i p2 a, nth i (sync v) None = Some a -> 
                ~(mLs TrM i) p2 (nthS i s2) \/ ~Trans (Li i) a (nthS i p1) p2))
        \/ (NetTransSL M v s2 s1 /\ 
             (forall i p2 a, nth i (sync v) None = Some a -> 
                ~(mLs TrM i) p2 (nthS i s2) \/ ~Trans (Li i) a p2 (nthS i p1))) )
        -> forall i, I i -> Ac v i -> Init (Li i) (nthS i p1).
Proof.
intros n M RS TrM I s1 s2 p1 v Mp Li HMp1 HTM i HIi HAci.
destruct HAci as [a Ha].
destruct HMp1 as [HSs1 [HSp1 Hmpi]].
specialize Hmpi with i.
assert (i < n) as Hin by (apply (listn_nth_bound None (sync v) i) ; congruence).
destruct HTM as [[[HVv HT] HTM] | [[HVv HT] HTM]] ;
specialize HT with i ; rewrite Ha in HT ;
apply NNPP ; 
unfold not ; intros ;
pose proof (D_cond _ _ (mLs TrM i) (nthS i s1) (nthS i s2) (nthS i p1) a) ;
destruct (m_states _ _ (mLs TrM i) _ _ (Hmpi Hin HIi)) as [HSpi _] ;
destruct (trans_cond _ _ _ _ (HT Hin)) as [HSs1i [HSs2i _]] ;
intuition ;
destruct H3 as [p2i [HSp2i [HTL Hmp2i]]] ;
specialize HTM with i p2i a ;
intuition.
Qed.

Lemma MapR_D_reverse :
  forall {n} {M} {RS} (TrM : @NetTransform n M RS) I s1 s2 p1 v,
    let Mp := MapR TrM I in
    let Ri (i : nat) := R (nth i (Rules RS) d_tr) in
      Mp p1 s1 ->
      (    (NetTransSL TrM v s1 s2 /\
             (forall i p2 a, nth i (sync v) None = Some a -> 
                ~(mR _ (TrG TrM i) p2 (nthS i s2)) \/ ~Trans (Ri i) a (nthS i p1) p2))
        \/ (NetTransSL TrM v s2 s1 /\ 
             (forall i p2 a, nth i (sync v) None = Some a -> 
                ~(mR _ (TrG TrM i) p2 (nthS i s2)) \/ ~Trans (Ri i) a p2 (nthS i p1))) )
        -> forall i, I i -> Ac v i -> Init (Ri i) (nthS i p1).
Proof.
intros n M RS TrM I s1 s2 p1 v Mp Ri HMp1 HTTrM i HIi HAci.
destruct HAci as [a Ha].
destruct HMp1 as [HSs1 [HSp1 Hmpi]].
specialize Hmpi with i.
assert (i < n) as Hin by (apply (listn_nth_bound None (sync v) i) ; congruence).
destruct HTTrM as [[[HVv HT] HTTrM] | [[HVv HT] HTTrM]] ;
specialize HT with i ; rewrite Ha in HT ;
apply NNPP ; 
unfold not ; intros ;
pose proof (D_cond _ _ (mR _ (TrG TrM i)) (nthS i s1) (nthS i s2) (nthS i p1) a) ;
destruct (m_states _ _ (mR _ (TrG TrM i)) _ _ (Hmpi Hin HIi)) as [HSpi _] ;
destruct (trans_cond _ _ _ _ (HT Hin)) as [HSs1i [HSs2i _]] ;
rewrite Procs_cond in HSs2i, HSs1i, HT ;
intuition ;
destruct H3 as [p2i [HSp2i [HTL Hmp2i]]] ;
specialize HTTrM with i p2i a ;
intuition.
Qed.

Lemma exists_mapState_rewrite_helper :
  forall n (P : Prop) (Q R S : nat -> State n -> Prop),
    (P /\ exists q, forall i, i < n -> Q i q /\ R i q /\ S i q)
      -> exists q, (P /\ (forall i, i < n -> Q i q) /\ (forall i, i < n -> R i q))
         /\ (forall i, i < n -> S i q).
Proof.
intuition ; destruct H1 ; exists x.
intuition ; specialize H with i ; intuition.
Qed.

Lemma exists_mapLState :
  forall {n}{M}{RS}(TrM : @NetTransform n M RS) (I : Ensemble nat) s,
    States (SystemLTS M) s -> (forall i, I i -> MatchLIds TrM s i) -> 
      exists q, MapL TrM I q s /\ forall i, i < n -> ~I i -> Init d_lts (nthS i q).
Proof.
unfold MapL.
simpl.
intros.
apply exists_mapState_rewrite_helper.
split ; try assumption.
set (P := fun a b => 
  States (nth a (Procs (LeftNetwork RS I)) d_lts) (Bld_list1 b) /\
  (I a -> (mLs TrM a) (Bld_list1 b) (nthS a s)) /\
  (~ I a -> SingletonS d_state (Bld_list1 b))).
apply (listn_forall_exists _ n dummy P).
unfold P ; clear P.
intros.
destruct (EnsembleNat_dec I i) as [HI | HI].
(* i in I *)
rewrite (LeftNetwork_Procs_i_in_I RS I) ; try assumption.
destruct (H0 i) as [_ [qi Hqi]] ; try exact HI.
exists (hd dummy qi).
rewrite list1_prop.
intuition.
destruct (m_states _ _ _ _ _ Hqi) as [HSqi _].
exact HSqi.
(* i not in I *)
exists dummy.
unfold LeftNetwork ; rewrite (FilterNetwork_Procs_i_not_in_I _ I) ; try assumption.
simpl.
unfold SingletonS.
intuition.
Qed.

Lemma exists_mapRState :
  forall {n}{M}{RS}(TrM : @NetTransform n M RS) (I : Ensemble nat) s,
    States (SystemLTS TrM) s -> (forall i, I i -> MatchRIds TrM s i) -> 
      exists q, MapR TrM I q s /\ forall i, i < n -> ~I i -> Init d_lts (nthS i q).
Proof.
unfold MapR.
simpl.
intros.
apply exists_mapState_rewrite_helper.
split ; try assumption.
set (P := fun a b => 
  States (nth a (Procs (RightNetwork RS I)) d_lts) (Bld_list1 b) /\
  (I a -> (mR _ (TrG TrM a)) (Bld_list1 b) (nthS a s)) /\
  (~ I a -> SingletonS d_state (Bld_list1 b))).
apply (listn_forall_exists _ n dummy P).
unfold P ; clear P.
intros.
destruct (EnsembleNat_dec I i) as [HI | HI].
(* i in I *)
rewrite (RightNetwork_Procs_i_in_I RS I) ; try assumption.
destruct (H0 i) as [_ [qi Hqi]] ; try exact HI.
exists (hd dummy qi).
rewrite list1_prop.
intuition.
destruct (m_states _ _ _ _ _ Hqi) as [HSqi _].
exact HSqi.
(* i not in I *)
exists dummy.
unfold RightNetwork ; rewrite (FilterNetwork_Procs_i_not_in_I _ I) ; try assumption.
simpl.
unfold SingletonS.
intuition.
Qed.

(* ==================================================================
     Network transformation properties
   ================================================================== *)

Lemma NotAcMatchL_prop :
  forall {n} {M} {RS} (TrM : @NetTransform n M RS) v s s',
    NetTransSL M v s s' -> forall i, i < n -> ~Ac v i -> MatchLIds TrM s i = MatchLIds TrM s' i.
Proof.
intros.
unfold MatchLIds, MatchIds.
rewrite (NetTransSL_i_not_in_Ac _ v s s' H i H0 H1).
reflexivity.
Qed.

Lemma NotAcMatchR_prop :
  forall {n} {M} {RS} (TrM : @NetTransform n M RS) v s s',
    NetTransSL TrM v s s' -> forall i, i < n -> ~Ac v i -> MatchRIds TrM s i = MatchRIds TrM s' i.
Proof.
intros.
unfold MatchRIds, MatchIds.
rewrite (NetTransSL_i_not_in_Ac _ v s s' H i H0 H1).
reflexivity.
Qed.

Ltac TransMatch_prop :=
  intros n M RS TrM I v s s' sm s'm HMs HMs' HT HT' ;
  apply Extensionality_Ensembles ;
  unfold Same_set, Included, In ;
  set (HMsm := sm) ; clear HMsm ;
  split ; intros x HM ; [ destruct HMs' as [_ [_ HMsm]] | destruct HMs as [_ [_ HMsm]]] ;
  specialize HMsm with x ;
  destruct HM as [Hxn Hp] ;
  split ; try exact Hxn ;
  destruct (EnsembleNat_dec I x) as [HI | HI] ;
  try first 
  [ (* I x *)
    pose proof (m_states _ _ _ _ _ (HMsm Hxn HI)) ;
    solve [ exists (nthS x s'm) ; intuition | 
            exists (nthS x sm) ; intuition ]
  | (* ~I x *)
    pose proof ((FilterNetwork_Trans_Ac_in_I _ I v sm s'm HT' x)) as HAc ;
    apply contrapositive in HAc ; try exact HI ;
    rewrite <- (NetTransSL_i_not_in_Ac _ _ _ _ HT x Hxn HAc) 
    || rewrite (NetTransSL_i_not_in_Ac _ _ _ _ HT x Hxn HAc) ;
    exact Hp
  ].

Lemma Trans_MatchL_prop : 
  forall {n} {M} {RS} (TrM : @NetTransform n M RS) I v s s' sm s'm,
    MapL TrM I sm s -> MapL TrM I s'm s' -> NetTransSL M v s s' -> 
      NetTransSL (LeftNetwork RS I) v sm s'm -> MatchLIds TrM s = MatchLIds TrM s'.
Proof.
TransMatch_prop.
Qed.

Lemma Trans_MatchR_prop : 
  forall {n} {M} {RS} (TrM : @NetTransform n M RS) I v s s' sm s'm,
    MapR TrM I sm s -> MapR TrM I s'm s' -> NetTransSL TrM v s s' -> 
      NetTransSL (RightNetwork RS I) v sm s'm -> MatchRIds TrM s = MatchRIds TrM s'.
Proof.
TransMatch_prop.
Qed.

Lemma TauPath_MatchL_prop : 
  forall {n} {M} {RS} (TrM : @NetTransform n M RS) I s s' sm s'm,
    MapL TrM I sm s -> MapL TrM I s'm s' ->  (forall i, i < n -> ~I i -> nthS i s = nthS i s') ->
      TauPath (SystemLTS (LeftNetwork RS I)) sm s'm -> MatchLIds TrM s = MatchLIds TrM s'.
Proof.
intros n M RS TrM I s s' sm s'm H H0 H1 H2.
apply Extensionality_Ensembles.
unfold image_mr, Same_set, Included, In.
split ; intros x H3 ;
destruct H3 as [Hxn Hm] ;
destruct (EnsembleNat_dec I x) as [HI | HI] ; split ; try exact Hxn ;
  [ (* x in I *)
    exists (nthS x s'm) ;
    destruct H0 as [_ [HSsm HMsm]]
  | (* x not in I *) rewrite <- (H1 x Hxn HI)
  | (* x in I *)
    exists (nthS x sm) ;
    destruct H as [_ [HSsm HMsm]] 
  | (* x not in I *) rewrite (H1 x Hxn HI)
  ] ; (* x not in I *) try exact Hm ;
      (* x in I *) apply HMsm ; assumption.
Qed.

Lemma TauPath_MatchR_prop : 
  forall {n} {M} {RS} (TrM : @NetTransform n M RS) I s s' sm s'm,
    MapR TrM I sm s -> MapR TrM I s'm s' ->  (forall i, i < n -> ~I i -> nthS i s = nthS i s') ->
      TauPath (SystemLTS (RightNetwork RS I)) sm s'm -> MatchRIds TrM s = MatchRIds TrM s'.
Proof.
intros n M RS TrM I s s' sm s'm H H0 H1 H2.
apply Extensionality_Ensembles.
unfold image_mr, Same_set, Included, In.
split ; intros x H3 ;
destruct H3 as [Hxn Hm] ;
destruct (EnsembleNat_dec I x) as [HI | HI] ; split ; try exact Hxn ;
  [ (* x in I *)
    exists (nthS x s'm) ;
    destruct H0 as [_ [HSsm HMsm]]
  | (* x not in I *) rewrite <- (H1 x Hxn HI)
  | (* x in I *)
    exists (nthS x sm) ;
    destruct H as [_ [HSsm HMsm]] 
  | (* x not in I *) rewrite (H1 x Hxn HI)
  ] ; (* x not in I *) try exact Hm ;
      (* x in I *) apply HMsm ; assumption.
Qed.

Ltac MatchedIds_not_prop :=
  intros n M RS TrM s i H H0 p ;
  unfold not ; intro H1 ;
  apply H0 ;
  split ; try exact H ;
  pose proof (m_states _ _ _ _ _ H1) as [HSq HSs] ;
  exists p ;
  intuition.

Lemma MatchedLIds_not_prop :
  forall {n} {M} {RS} (TrM : @NetTransform n M RS) s i, i < n ->
    ~MatchLIds TrM s i -> forall p, ~(mLs TrM i) p (nthS i s).
Proof.
MatchedIds_not_prop.
Qed.

Lemma MatchedRIds_not_prop :
  forall {n} {M} {RS} (TrM : @NetTransform n M RS) s i, i < n ->
    ~MatchRIds TrM s i -> forall p, ~(mR _ (TrG TrM i)) p (nthS i s).
Proof.
MatchedIds_not_prop.
Qed.

Lemma ComposedToState_MatchLIds :
  forall {n}{M : LTSNetwork}{RS : RuleSystem}(TrM : @NetTransform n M RS)(v : @SyncLaw n) s s' p p',
    NetTransSL M v s s' -> p' = Cmps p s' (Ac v) -> (forall i, MatchLIds TrM s i = MatchRIds TrM p i) -> 
    (forall i, Ac v i -> forall q', (mLs TrM i) q' (nthS i s') -> Init (L (nth i (Rules RS) d_tr)) q') ->
      MatchLIds TrM s' = MatchRIds TrM p'.
Proof.
intros n M RS TrM v s s' p p' HT Heqp' HCIdsi HILAc'.
apply Extensionality_Ensembles.
unfold Same_set, Included, In.
assert (Trans (SystemLTS M) (act v) s s') as HT' by (exists v ; auto).
destruct (trans_cond _ _ _ _ HT') as [_ [HSs' _]].
pose proof (Cmps_In p s' (Ac v)) as Hp'.
rewrite <- Heqp' in Hp'.
split ; intros x Hx ; pose proof Hx as [Hxn _] ; destruct (Hp' x Hxn) as [Hp'Some Hp'None] ;
destruct (EnsembleNat_dec (Ac v) x) as [Hv | Hv] ;
  [ pose proof (HILAc' x Hv) as HIq' | | 
    pose proof (Ac_not_removed_InitR TrM v s' p p' x HSs' Heqp' Hv) as HIq' | ] ; solve
  [ (* Ac v x *)
    destruct Hx as [_ [q' Hmq']] ;
    pose proof (HIq' q' Hmq') as HIq'x ;
    try rewrite <- Init_L_to_Init_R in HIq'x ;
    split ; try exact Hxn ;
    exists q' ; unfold nthS in Hmq' |- * ;
    rewrite (Hp'Some Hv) || rewrite (Hp'Some Hv) in Hmq' ;
    rewrite <- (mR_init_cond _ (TrG TrM x) q' HIq'x)
    || rewrite <- (mR_init_cond _ (TrG TrM x) q' HIq'x) in Hmq' ;
    exact Hmq'
  | (* ~Ac v x *)
    unfold MatchLIds, MatchRIds, MatchIds in Hx, HCIdsi |- * ;
    rewrite <- (NetTransSL_i_not_in_Ac M _ _ _ HT x Hxn Hv) in Hx
    || rewrite <- (NetTransSL_i_not_in_Ac M _ _ _ HT x Hxn Hv) ;
    unfold nthS in Hx |- * ; rewrite (Hp'None Hv) || rewrite (Hp'None Hv) in Hx ;
    rewrite <- HCIdsi || rewrite <- HCIdsi in Hx ; exact Hx ].
Qed.

Lemma ComposedToState_MatchRIds :
  forall {n}{M : LTSNetwork}{RS : RuleSystem}(TrM : @NetTransform n M RS)(v : @SyncLaw n) s s' p p',
    NetTransSL TrM v s s' -> p' = Cmps p s' (Ac v) -> (forall i, MatchLIds TrM p i = MatchRIds TrM s i) -> 
    (forall i, Ac v i -> forall q', (mR _ (TrG TrM i)) q' (nthS i s') -> Init (R (nth i (Rules RS) d_tr)) q') ->
      MatchLIds TrM p' = MatchRIds TrM s'.
Proof.
intros n M RS TrM v s s' p p' HT Heqp' HCIdsi HILAc'.
apply Extensionality_Ensembles.
unfold Same_set, Included, In.
assert (Trans (SystemLTS TrM) (act v) s s') as HT' by (exists v ; auto).
destruct (trans_cond _ _ _ _ HT') as [_ [HSs' _]].
pose proof (Cmps_In p s' (Ac v)) as Hp'.
rewrite <- Heqp' in Hp'.
split ; intros x Hx ; pose proof Hx as [Hxn _] ; destruct (Hp' x Hxn) as [Hp'Some Hp'None] ;
destruct (EnsembleNat_dec (Ac v) x) as [Hv | Hv] ;
  [ pose proof (Ac_not_removed_InitL TrM v s' p p' x HSs' Heqp' Hv) as HIq' |
  | pose proof (HILAc' x Hv) as HIq' | ] ; try solve
  [ (* Ac v x *)
    destruct Hx as [_ [q' Hmq']] ;
    pose proof (HIq' q' Hmq') as HIq'x ;
    try rewrite <- Init_L_to_Init_R in HIq'x ;
    split ; try exact Hxn ;
    exists q' ; unfold nthS in Hmq' |- * ;
    rewrite (Hp'Some Hv) || rewrite (Hp'Some Hv) in Hmq' ;
    rewrite <- (mR_init_cond _ (TrG TrM x) q' HIq'x)
    || rewrite <- (mR_init_cond _ (TrG TrM x) q' HIq'x) in Hmq' ;
    exact Hmq'
  | (* ~Ac v x *)
    unfold MatchLIds, MatchRIds, MatchIds in Hx, HCIdsi |- * ;
    rewrite <- (NetTransSL_i_not_in_Ac TrM _ _ _ HT x Hxn Hv) in Hx
    || rewrite <- (NetTransSL_i_not_in_Ac TrM _ _ _ HT x Hxn Hv) ;
    unfold nthS in Hx |- * ; rewrite (Hp'None Hv) || rewrite (Hp'None Hv) in Hx ;
    rewrite <- HCIdsi || rewrite <- HCIdsi in Hx ; exact Hx ].
Qed.



(* If a transition is only partly matched, the matched state(s) must be glue-states *)
Lemma TransPartlyMatchedL_prop : 
  forall {n} {M} {RS} (TrM : @NetTransform n M RS) v a s1 s2 p i,
    let Li := L (nth i (Rules RS) d_tr) in
      NetTransSL M v s1 s2 -> nth i (sync v) None = Some a ->
        (  mLs TrM i p (nthS i s1) /\ (forall p2, ~mLs TrM i p2 (nthS i s2) \/ ~Trans Li a p p2)
        \/ mLs TrM i p (nthS i s2) /\ (forall p1, ~mLs TrM i p1 (nthS i s1) \/ ~Trans Li a p1 p))
        -> Init Li p.
Proof.
intros.
assert (Trans (SystemLTS M) (act v) s1 s2) as HT' by (simpl ; exists v ; intuition).
assert (i < n) as Hin by (apply (listn_nth_bound None (sync v)) ; congruence).
destruct H as [Hv HT].
specialize HT with i.
rewrite H0 in HT.
destruct H1 as [[Hmp HNm] | [Hmp HNm]] ;
  apply (D_cond_reverse _ _ (mLs TrM i) _ _ p a (HT Hin)) ;
  [left | right] ;
  split ; assumption.
Qed.

Lemma TransPartlyMatchedR_prop : 
  forall {n} {M} {RS} (TrM : @NetTransform n M RS) v a s1 s2 p i,
    let Ri := R (nth i (Rules RS) d_tr) in
      NetTransSL TrM v s1 s2 -> nth i (sync v) None = Some a ->
        (  mR _ (TrG TrM i) p (nthS i s1) /\ (forall p2, ~mR _ (TrG TrM i) p2 (nthS i s2) \/ ~Trans Ri a p p2)
        \/ mR _ (TrG TrM i) p (nthS i s2) /\ (forall p1, ~mR _ (TrG TrM i) p1 (nthS i s1) \/ ~Trans Ri a p1 p))
        -> Init Ri p.
Proof.
intros.
assert (Trans (SystemLTS TrM) (act v) s1 s2) as HT' by (simpl ; exists v ; intuition).
assert (i < n) as Hin by (apply (listn_nth_bound None (sync v)) ; congruence).
destruct H as [Hv HT].
specialize HT with i.
rewrite H0 in HT.
rewrite Procs_cond in HT ; try exact Hin.
destruct H1 as [[Hmp HNm] | [Hmp HNm]] ;
  apply (D_cond_reverse _ _ (mR _ (TrG TrM i)) _ _ p a (HT Hin)) ;
  [left | right] ;
  split ; assumption.
Qed.

Lemma TransPartlyMatchedL_prop2 : forall {n} {M} {RS} (TrM : @NetTransform n M RS) v s s' p i,
  NetTransSL M v s s' -> Ac v i -> ((mLs TrM i) p (nthS i s) \/ (mLs TrM i) p (nthS i s'))->
  (forall j, j < n -> Ac v j -> 
    ~(exists q q', (mLs TrM j) q (nthS j s) /\ (mLs TrM j) q' (nthS j s') 
        /\ (exists aj, nth j (sync v) None = Some aj /\ Trans (L (nth j (Rules RS) d_tr)) aj q q')))
  -> Init (L (nth i (Rules RS) d_tr)) p.
Proof.
intros.
pose proof H0 as [a Ha].
assert (i < n) as Hin by ( apply (listn_nth_bound None (sync v)) ; congruence ).
apply (TransPartlyMatchedL_prop TrM _ _ _ _ p i H Ha).
destruct H1 ; [left | right] ;
split ; try exact H1 ;
apply forall_not_or_2_not_exists_and ;
unfold not ; intro ;
destruct H3 as [p' [Hp' HTp']] ;
apply (H2 i Hin H0) ;
[ exists p, p' | exists p', p ] ;
intuition ;
exists a ;
split ; assumption.
Qed.

Lemma TransPartlyMatchedR_prop2 : forall {n} {M} {RS} (TrM : @NetTransform n M RS) v s s' p i,
  NetTransSL TrM v s s' -> Ac v i -> (mR _ (TrG TrM i) p (nthS i s)) \/ (mR _ (TrG TrM i) p (nthS i s')) ->
  (forall j, j < n -> Ac v j -> 
    ~(exists q q', mR _ (TrG TrM j) q (nthS j s) /\ mR _ (TrG TrM j) q' (nthS j s') 
        /\ (exists aj, nth j (sync v) None = Some aj /\ Trans (R (nth j (Rules RS) d_tr)) aj q q')))
  -> Init (R (nth i (Rules RS) d_tr)) p.
Proof.
intros.
pose proof H0 as [a Ha].
assert (i < n) as Hin by ( apply (listn_nth_bound None (sync v)) ; congruence ).
apply (TransPartlyMatchedR_prop TrM _ _ _ _ p i H Ha).
destruct H1 ; [left | right] ;
split ; try exact H1 ;
apply forall_not_or_2_not_exists_and ;
unfold not ; intro ;
destruct H3 as [p' [Hp' HTp']] ;
apply (H2 i Hin H0) ;
[ exists p, p' | exists p', p ] ;
intuition ;
exists a ;
split ; assumption.
Qed.

Ltac Ac_Trans_MatchIds_prop :=
  intros n M RS TrM v s s' H i H0 H1 ;
  apply NNPP ; unfold not ; intro H2 ;
  apply AcNone_prop in H2 ;
  destruct H as [H4  H3] ;
  specialize H3 with i ;
  rewrite H2 in H3 ;
  destruct H3 as [HRW _] ; try exact H0 ;
  destruct H1 as [[HM HN] | [HN HM]] ;
  apply HN ;
  destruct HM as [_ HX] ;
  unfold nthS in HX ;
  [ rewrite HRW in HX | rewrite <- HRW in HX ] ;
  split ; assumption.

Lemma Ac_Trans_MatchLIds_prop : 
  forall {n} {M} {RS} (TrM : @NetTransform n M RS) v s s', NetTransSL M v s s' -> 
    forall i, i < n -> 
      (    (MatchLIds TrM s i /\ ~MatchLIds TrM s' i)
        \/ (~MatchLIds TrM s i /\ MatchLIds TrM s' i)
      ) -> Ac v i.
Proof.
Ac_Trans_MatchIds_prop.
Qed.

Lemma Ac_Trans_MatchRIds_prop : 
  forall {n} {M} {RS} (TrM : @NetTransform n M RS) v s s', NetTransSL TrM v s s' -> 
    forall i, i < n -> 
      (    (MatchRIds TrM s i /\ ~MatchRIds TrM s' i)
        \/ (~MatchRIds TrM s i /\ MatchRIds TrM s' i)
      ) -> Ac v i.
Proof.
Ac_Trans_MatchIds_prop.
Qed.

Lemma Ac_Matched_Trans_MatchLIds_prop : 
  forall {n} {M} {RS} (TrM : @NetTransform n M RS) I v s' sm s'm,
    MapL TrM I s'm s' -> NetTransSL (LeftNetwork RS I) v sm s'm \/ NetTransSL (LeftNetwork RS I) v s'm sm -> 
      forall i, i < n -> ~MatchLIds TrM s' i -> ~Ac v i.
Proof.
unfold not.
intros n M RS TrM I v s' sm s'm HMs'm HT i Hin HNMs' HAci.
apply HNMs'.
split ; try exact Hin.
destruct HMs'm as [HSs' [HSs'm Hms'm]].
specialize Hms'm with i.
destruct HT as [HT | HT] ; 
pose proof (FilterNetwork_Trans_Ac_in_I _ _ _ _ _ HT i HAci) ;
exists (nthS i s'm) ;
apply Hms'm ; assumption.
Qed.

Lemma Ac_Matched_Trans_MatchRIds_prop : 
  forall {n} {M} {RS} (TrM : @NetTransform n M RS) I v s' sm s'm,
    MapR TrM I s'm s' -> NetTransSL (RightNetwork RS I) v sm s'm \/ NetTransSL (RightNetwork RS I) v s'm sm ->  
      forall i, i < n -> ~MatchRIds TrM s' i -> ~Ac v i.
Proof.
unfold not.
intros n M RS TrM I v s' sm s'm HMs'm HT i Hin HNMs' HAci.
apply HNMs'.
split ; try exact Hin.
destruct HMs'm as [HSs' [HSs'm Hms'm]].
specialize Hms'm with i.
destruct HT as [HT | HT] ;
pose proof (FilterNetwork_Trans_Ac_in_I _ _ _ _ _ HT i HAci) ;
exists (nthS i s'm) ;
apply Hms'm ; assumption.
Qed.

End TransformationNetwork.